from __future__ import annotations

__doc__ = """
.. autoclass:: TuningParameter
.. autoclass:: IntParameter
.. autoclass:: BoolParameter
.. autoclass:: TupleParameter

.. autofunction:: autotune
.. autofunction:: transform_param
.. autofunction:: einsum_arg

.. class:: ConvertibleToTuningParamT

    A type alias for ``Union[TuningParameter, Tuple[Any, ...]]``.
    See :attr:`transform_param.func`.
"""
import abc
import logging
import os
import sqlite3
from dataclasses import dataclass
from functools import cache, cached_property
from typing import (
    TYPE_CHECKING,
    Any,
)

import numpy as np
import opentuner
from immutables import Map

from feinsum.einsum import INT_CLASSES, BatchedEinsum, IntegralT, ShapeComponentT
from feinsum.sql_utils import DEFAULT_DB, TIMINGS_TABLENAME

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence

    import loopy as lp
    import pyopencl as cl

    from feinsum.typing import TransformT


logger = logging.getLogger(__name__)


# {{{ supported tuning parameters


class TuningParameter(abc.ABC):  # noqa: B024
    """
    Records the parameter space of a code-transformation.

    .. note::

        This is an abstract class.
    """


@dataclass(frozen=True, init=False)
class IntParameter(TuningParameter):
    """
    A parameter that takes values in the range ``[low, high]``.
    """

    low: IntegralT
    high: IntegralT

    def __init__(self, low: ShapeComponentT, high: ShapeComponentT):
        if not isinstance(low, INT_CLASSES):
            raise ValueError("low must be an integer")

        if not isinstance(high, INT_CLASSES):
            raise ValueError("high must be an integer")

        object.__setattr__(self, "low", low)
        object.__setattr__(self, "high", high)


@dataclass(frozen=True)
class BoolParameter(TuningParameter):
    """
    A parameter that can be either *True* or *False*.
    """


@dataclass(frozen=True)
class TupleParameter(TuningParameter):
    """
    A tuple of parameters. The resulting parameter space is a Cartesian product of
    the individual elements of the tuple.
    """

    _data: tuple[TuningParameter, ...]


# }}}


ConvertibleToTuningParamT = tuple[Any, ...] | TuningParameter


# {{{ einsum_arg/transform_param


@dataclass(frozen=True, repr=True)
class einsum_arg:  # noqa: N801
    """
    Decorate to a template transformation to inform
    :func:`autotune` about a static argument to the transformation
    implementation.

    :param arg_name: Name of the argument in the template transform
        that is to be statically set.
    :param param_getter: A callable that expects an einsum and returns the
        value of the static argument.
    """

    var_name: str
    func: Callable[[BatchedEinsum], Any]

    def __call__(self, fn: Callable[..., Any]) -> ParametrizedTransform:
        if isinstance(fn, ParametrizedTransform):
            return ParametrizedTransform(
                fn.transform,
                (self, *fn.einsum_derivative_args),
                fn.transform_params,
            )
        else:
            from functools import cache

            return ParametrizedTransform(cache(fn), (self,), ())


@dataclass(frozen=True, repr=True)
class transform_param:  # noqa: N801
    """
    Decorate to a template transformation to inform
    :func:`autotune` about the parameter space.

    :param arg_name: Name of the argument in the template
        transformation to be parametrized.
    :param func: A callable that expects an einsum and returns the
        parameter space for that einsum. The returned parameter space could be an
        instance of:
        - A :class:`TuningParameter`.
        - A :class:`tuple` of :class:`ConvertibleToTuningParamT` types that is
          (internally) mapped to a :class:`TupleParameter`.
    """

    var_name: str
    func: Callable[[BatchedEinsum], ConvertibleToTuningParamT]

    def __call__(
        self, fn: Callable[..., lp.TranslationUnit]
    ) -> ParametrizedTransform:
        if isinstance(fn, ParametrizedTransform):
            return ParametrizedTransform(
                fn.transform,
                fn.einsum_derivative_args,
                (self, *fn.transform_params),
            )
        else:
            from functools import cache

            return ParametrizedTransform(cache(fn), (), (self,))


@dataclass(frozen=True, repr=True)
class ParametrizedTransform:
    transform: Callable[..., lp.TranslationUnit]
    einsum_derivative_args: tuple[einsum_arg, ...]
    transform_params: tuple[transform_param, ...]

    def __call__(self, *args: Any, **kwargs: Any) -> lp.TranslationUnit:
        return self.transform(*args, **kwargs)

    def bind_args(self, einsum: BatchedEinsum, **transform_args: Any) -> TransformT:
        """
        Binds *transform_args* to *self* and returns a python callable
        to the corresponding instance in the self space.
        """
        from functools import partial

        py_clbl = partial(
            self.transform,
            **{
                arg.var_name: arg.func(einsum) for arg in self.einsum_derivative_args
            },
            **transform_args,
        )
        return py_clbl


# }}}


@cache
def _get_impls_path() -> str:
    import importlib.util

    feinsum_spec = importlib.util.find_spec("feinsum")
    assert feinsum_spec is not None
    assert feinsum_spec.origin is not None

    return os.path.abspath(
        os.path.join(feinsum_spec.origin, os.path.pardir, "tuning", "impls")
    )


class ConfigurationNotInDBError(LookupError):
    pass


def get_transform_func_from_module_path(module_path: str) -> ParametrizedTransform:
    from importlib import util

    _, filename = os.path.split(module_path)

    assert filename.endswith(".py")

    spec = util.spec_from_file_location(filename[:-3], module_path)
    if spec is None or spec.loader is None:
        raise RuntimeError(
            "Could not import 'transform' function" f" from {module_path}."
        )
    transform_module = util.module_from_spec(spec)
    spec.loader.exec_module(transform_module)
    transform_obj = transform_module.transform

    if isinstance(transform_obj, ParametrizedTransform):
        return transform_obj
    else:
        assert callable(transform_obj)
        return ParametrizedTransform(transform_obj, (), ())


def _convert_to_tuning_param(param: ConvertibleToTuningParamT) -> TuningParameter:
    if isinstance(param, TuningParameter):
        return param
    elif isinstance(param, tuple):
        return TupleParameter(tuple(_convert_to_tuning_param(el) for el in param))
    else:
        raise ValueError(
            "Only instances of ConvertibleToTuningParamT" " are supported."
        )


def _get_opentuner_param_name(key: tuple[str, ...]) -> str:
    return "_".join(key)


def _get_opentuner_params_from_tuning_param(
    key: tuple[str, ...], tuning_param: TuningParameter
) -> frozenset[opentuner.manipulator.Parameter]:
    if isinstance(tuning_param, IntParameter):
        from opentuner.search.manipulator import IntegerParameter

        return frozenset(
            [
                IntegerParameter(
                    _get_opentuner_param_name(key),
                    tuning_param.low,
                    tuning_param.high,
                )
            ]
        )
    elif isinstance(tuning_param, BoolParameter):
        from opentuner.search.manipulator import BooleanParameter

        return frozenset([BooleanParameter(_get_opentuner_param_name(key))])
    elif isinstance(tuning_param, TupleParameter):
        from functools import reduce

        return reduce(
            frozenset.union,
            (
                _get_opentuner_params_from_tuning_param((*key, f"_fetup_{i}"), k)
                for i, k in enumerate(tuning_param._data)
            ),
            frozenset(),
        )
    elif isinstance(tuning_param, TuningParameter):
        raise NotImplementedError(type(tuning_param))
    else:
        raise TypeError(type(tuning_param))


def _reconstruct_transform_params_from_opentuner_config(
    config: Mapping[str, Any],
    transform_params: tuple[transform_param, ...],
    ensm: BatchedEinsum,
) -> Mapping[str, Any]:

    def rec(key: tuple[str, ...], tuning_param: TuningParameter) -> Any:
        if isinstance(tuning_param, (IntParameter, BoolParameter)):
            param_name = _get_opentuner_param_name(key)
            return config[param_name]
        elif isinstance(tuning_param, TupleParameter):
            return tuple(
                rec((*key, f"_fetup_{isubparam}"), subparam)
                for isubparam, subparam in enumerate(tuning_param._data)
            )
        elif isinstance(tuning_param, TuningParameter):
            raise NotImplementedError(type(tuning_param))
        else:
            raise TypeError(type(tuning_param))

    return {
        param.var_name: rec(
            (param.var_name,), _convert_to_tuning_param(param.func(ensm))
        )
        for param in transform_params
    }


def _get_opentuner_config_from_transform_config(
    config: Mapping[str, Any],
) -> Map[str, Any]:
    result = {}

    def rec(key: tuple[str, ...], param: Any) -> None:
        if isinstance(param, (INT_CLASSES, bool)):
            result[_get_opentuner_param_name(key)] = param
            return
        elif isinstance(param, tuple):
            for ipar, par in enumerate(param):
                rec((*key, f"_fetup_{ipar}"), par)
        else:
            raise NotImplementedError(type(param))

    for k, v in config.items():
        rec((k,), v)

    return Map(result)


# {{{ Opentuner entrypoint


# type-ignored as we are sub-classing Any type.
class OpentunerTuner(opentuner.MeasurementInterface):  # type: ignore[misc]
    def __init__(
        self,
        args: Any,
        einsum: BatchedEinsum,
        cl_ctx: cl.Context,
        module_path: str,
        long_dim_length: int,
        db_path: str,
        *,
        # Args to super class ->
        project_name: str | None = None,
        program_name: str = "unknown",
        program_version: str = "unknown",
        manipulator: Any | None = None,
        objective: Any | None = None,
        input_manager: Any | None = None,
    ) -> None:
        from feinsum.canonicalization import canonicalize_einsum

        super().__init__(
            args=args,
            project_name=project_name,
            program_name=program_name,
            program_version=program_version,
            manipulator=manipulator,
            objective=objective,
            input_manager=input_manager,
        )

        self.einsum = canonicalize_einsum(einsum)
        self.cl_ctx = cl_ctx
        self.module_path = module_path
        self.long_dim_length = long_dim_length
        self.db_path = db_path

    @cached_property
    def transform_space_id(self) -> str:
        dirpath, filepath = os.path.split(self.module_path)

        if dirpath == _get_impls_path():
            return filepath
        else:
            return self.module_path

    @cached_property
    def conn(self) -> sqlite3.Connection:
        db = sqlite3.connect(self.db_path)
        from feinsum.sql_utils import _create_timings_table_if_non_existent

        _create_timings_table_if_non_existent(db)
        return db

    @cached_property
    def transform_func(self) -> ParametrizedTransform:
        return get_transform_func_from_module_path(self.module_path)

    def manipulator(self) -> opentuner.ConfigurationManipulator:
        from opentuner import ConfigurationManipulator

        manipulator = ConfigurationManipulator()

        for param in self.transform_func.transform_params:
            unprocessed_tuning_param = param.func(self.einsum)
            tuning_param = _convert_to_tuning_param(unprocessed_tuning_param)
            for opentuner_param in _get_opentuner_params_from_tuning_param(
                (param.var_name,), tuning_param
            ):
                manipulator.add_parameter(opentuner_param)

        return manipulator

    def seed_configurations(self) -> Sequence[Mapping[str, Any]]:
        from feinsum.sql_utils import (
            dump_arg_names,
            dump_arg_to_dtype,
            dump_device_name,
            dump_index_to_length,
            dump_op_info,
            load_transform_params,
        )

        cursor = self.conn.cursor()
        subscripts = self.einsum.get_subscripts()
        index_to_length = dump_index_to_length(self.einsum)
        arg_names = dump_arg_names(self.einsum)
        op_info = dump_op_info(self.einsum, self.long_dim_length)
        arg_to_dtype = dump_arg_to_dtype(self.einsum)
        (dev,) = self.cl_ctx.devices
        device_name = dump_device_name(dev)

        cursor.execute(
            " SELECT"
            "     transform_params"
            "  FROM "
            f"    {TIMINGS_TABLENAME}"
            " WHERE ("
            "    transform_id = ?"
            "    AND subscripts = ?"
            "    AND index_to_length = ?"
            "    AND args = ?"
            "    AND arg_to_dtype = ?"
            "    AND giga_op_info = ?"
            "    AND device_name = ?"
            ");",
            (
                self.transform_space_id,
                subscripts,
                index_to_length,
                arg_names,
                arg_to_dtype,
                op_info,
                device_name,
            ),
        )

        return [
            dict(
                _get_opentuner_config_from_transform_config(
                    load_transform_params(transform_params[0])
                )
            )
            for transform_params in cursor.fetchall()
        ]

    def query_from_db(self, parameters: Mapping[str, Any]) -> float:
        import json

        from feinsum.sql_utils import (
            dump_arg_names,
            dump_arg_to_dtype,
            dump_device_name,
            dump_index_to_length,
            dump_op_info,
        )

        cursor = self.conn.cursor()
        subscripts = self.einsum.get_subscripts()
        index_to_length = dump_index_to_length(self.einsum)
        arg_names = dump_arg_names(self.einsum)
        arg_to_dtype = dump_arg_to_dtype(self.einsum)
        transform_params_str = json.dumps(parameters, sort_keys=True)
        op_info = dump_op_info(self.einsum, self.long_dim_length)
        (dev,) = self.cl_ctx.devices
        device_name = dump_device_name(dev)

        cursor.execute(
            " SELECT"
            "     runtime_in_sec"
            "  FROM "
            f"    {TIMINGS_TABLENAME}"
            " WHERE ("
            "    subscripts = ?"
            "    AND index_to_length = ?"
            "    AND args = ?"
            "    AND arg_to_dtype = ?"
            "    AND transform_params = ?"
            "    AND giga_op_info = ?"
            "    AND device_name = ?"
            ");",
            (
                subscripts,
                index_to_length,
                arg_names,
                arg_to_dtype,
                transform_params_str,
                op_info,
                device_name,
            ),
        )
        stored_results: list[Sequence[float]] = cursor.fetchall()

        if not stored_results:
            raise ConfigurationNotInDBError
        else:
            return min(stored_result[0] for stored_result in stored_results)

    def record_into_db(self, runtime: float, parameters: Mapping[str, Any]) -> None:
        from feinsum.sql_utils import record_into_db

        record_into_db(
            self.einsum,
            self.cl_ctx,
            self.module_path,
            parameters,
            self.conn,
            self.long_dim_length,
        )

    def run(
        self, desired_result: opentuner.DesiredResult, input: Any, limit: Any
    ) -> opentuner.Result:
        from feinsum.diagnostics import InvalidParameterError
        from feinsum.measure import stringify_comparison_vs_roofline, timeit

        cfg = _reconstruct_transform_params_from_opentuner_config(
            desired_result.configuration.data,
            self.transform_func.transform_params,
            self.einsum,
        )

        logger.info(cfg)

        # {{{ query from DB

        try:
            result = self.query_from_db(cfg)
        except ConfigurationNotInDBError:
            pass
        else:
            logger.info("DB Hit")
            return opentuner.Result(time=result)

        # }}}

        bound_transform = self.transform_func.bind_args(self.einsum, **cfg)

        try:
            logger.info(
                "\n"
                + stringify_comparison_vs_roofline(
                    self.einsum,
                    transform=bound_transform,
                    cl_ctx=self.cl_ctx,
                    long_dim_length=self.long_dim_length,
                )
            )
        except InvalidParameterError as err:
            logger.info(f"Ignored configuration due to '{err}'.")
            return opentuner.Result(time=np.inf)

        runtime = timeit(
            self.einsum,
            cl_ctx=self.cl_ctx,
            transform=bound_transform,
            long_dim_length=self.long_dim_length,
        )
        self.record_into_db(runtime, cfg)

        return opentuner.Result(time=runtime)


# }}}


def autotune(
    einsum: BatchedEinsum,
    module_path: str,
    cl_ctx: cl.Context,
    *,
    db_path: str | None = None,
    long_dim_length: int = 100_000,
    stop_after: int | None = None,
) -> None:
    """
    For a transform space specified in *module_path*, searches the parameter
    space and records the timing results for each run in *db_path*.

    :param stop_after: After these many trials the routine exits. Pass *None*
        to go on indefinitely.
    """
    if not os.path.isabs(module_path):
        raise ValueError("autotune expects an absolute path for the module")

    if db_path is None:
        db_path = DEFAULT_DB

    from argparse import Namespace  # :puke: but required by opentuner. Big brain.

    kwargs: Mapping[str, Any] = {
        "machine_class": None,
        "parallel_compile": False,
        "test_limit": None,
        "stop_after": stop_after,
        "parallelism": 4,
        "pipelining": 0,
        "bail_threshold": 100,
        "no_dups": True,
        "seed_configuration": [],
        "results_log": None,
        "results_log_details": None,
        "quiet": False,
        "display_frequency": 10,
        "technique": None,
        "list_techniques": False,
        "generate_bandit_technique": False,
        "label": None,
        "print_search_space_size": False,
        "database": None,
        "print_params": False,
    }

    OpentunerTuner.main(
        args=Namespace(**kwargs),
        einsum=einsum,
        cl_ctx=cl_ctx,
        module_path=module_path,
        db_path=db_path,
        long_dim_length=long_dim_length,
    )


# vim: fdm=marker



"""
:mod:`loopy` helpers. The aim is to keep this module as lean as possible by
upstreaming the heavy lifting parts to :mod:`loopy` itself.

.. autofunction:: get_a_matched_einsum
.. autofunction:: match_t_unit_to_einsum
.. autofunction:: get_call_ids
.. autofunction:: hoist_invariant_multiplicative_terms_in_sum_reduction
.. autofunction:: extract_multiplicative_terms_in_sum_reduction_as_subst
"""

from collections.abc import Callable, Collection, Iterable, Mapping, Sequence
from typing import (
    Any,
    TypeVar,
    cast,
)

import islpy as isl
import loopy as lp
import numpy as np
import pymbolic.primitives as p
from bidict import frozenbidict
from immutables import Map
from loopy.diagnostic import LoopyError
from loopy.kernel import LoopKernel
from loopy.kernel.data import SubstitutionRule
from loopy.symbolic import CombineMapper, IdentityMapper, Reduction
from loopy.translation_unit import for_each_kernel
from more_itertools import zip_equal as szip
from pytools import memoize_on_first_arg

from feinsum.diagnostics import EinsumTunitMatchError
from feinsum.einsum import (
    BatchedEinsum,
    IntegralT,
    ShapeComponentT,
    ShapeT,
    SizeParam,
)

# {{{ matching loopy kernel against a ref. einsum


class ReductionCollector(CombineMapper[frozenset[p.Expression], []]):
    """
    A mapper that collects all instances of :class:`loopy.symbolic.Reduction`
    in an expression.
    """

    def combine(
        self, values: Iterable[frozenset[p.Expression]]
    ) -> frozenset[p.Expression]:
        from functools import reduce

        return reduce(frozenset.union, values, frozenset())

    def map_reduction(self, expr: Reduction) -> frozenset[p.Expression]:
        return self.combine([frozenset([expr]), super().map_reduction(expr)])

    def map_algebraic_leaf(self, expr: Any) -> frozenset[p.Expression]:
        return frozenset()

    def map_constant(self, expr: Any) -> frozenset[p.Expression]:
        return frozenset()


def _get_indices_from_assignee(assignee: p.Expression) -> tuple[str, ...]:
    r"""
    Returns the accessed indices in assignee. Expects the assignee to be seen in a
    batched-einsum matchable :class:`~loopy.LoopKernel`\ 's expression.
    """
    if isinstance(assignee, p.Variable):
        return ()
    elif isinstance(assignee, p.Subscript) and all(
        isinstance(idx, p.Variable) for idx in assignee.index_tuple
    ):
        return tuple(cast("p.Variable", idx).name for idx in assignee.index_tuple)
    else:
        raise EinsumTunitMatchError(
            "Not all instructions assign to Variable or"
            " a Subscript with free indices --"
            " not allowed."
        )


def _get_iname_length(
    kernel: lp.LoopKernel,
    iname: str,
    long_dim_length: IntegralT,
    iname_to_index: Mapping[str, str],
) -> ShapeComponentT:
    r"""
    Returns :math:`\mathcal{L}(\text{iname})`. In order to have the same iteration
    domain as that of an einsum, we enforce that *iname* has a zero lower bound
    (inclusive). In the case when :math:`\mathcal{L}(\text{iname}) \geq
    \text{long\_dim\_length}` or when the *iname*'s domain size is parametric we
    return :data:`numpy.inf` to denote an :math:`\infty`-long dimension.
    """
    from loopy.isl_helpers import static_max_of_pw_aff, static_min_of_pw_aff

    bounds = kernel.get_iname_bounds(iname)
    lbound_pwaff = bounds.lower_bound_pw_aff
    static_min_lbound_pwaff = static_min_of_pw_aff(lbound_pwaff, constants_only=True)
    if static_min_lbound_pwaff.get_constant_val().to_python() != 0:
        raise EinsumTunitMatchError(
            f"Iname {iname} that appears as an einsum index"
            f" in {kernel.name} does not have '0' as its"
            " lower bound -> not allowed in feinsum"
            " grammar."
        )

    ubound_pw_aff = bounds.upper_bound_pw_aff
    ubound = static_max_of_pw_aff(ubound_pw_aff, constants_only=False)

    if ubound.is_cst():
        ubound_val = ubound.get_pieces()[0][1].get_constant_val().to_python()
        assert isinstance(ubound_val, int)
        if ubound_val >= long_dim_length:
            assert iname_to_index[iname].islower()
            return SizeParam(iname_to_index[iname].upper())
        else:
            return ubound_val + 1
    else:
        assert iname_to_index[iname].islower()
        return SizeParam(iname_to_index[iname].upper())


@memoize_on_first_arg
def _infer_lp_types(t_unit: lp.TranslationUnit) -> lp.TranslationUnit:
    return lp.infer_unknown_types(t_unit)


def _get_dtype_for_subst_argument(
    t_unit: lp.TranslationUnit, kernel_name: str, rule_name: str
) -> np.dtype[Any]:
    """
    Returns the inferred data type for *rule_name*'s expression when all the
    arguments passed to it are equal to it kernel's
    :attr:`loopy.LoopKernel.index_dtype`.
    """
    from loopy.symbolic import SubstitutionMapper, SubstitutionRuleExpander
    from loopy.type_inference import TypeReader
    from pymbolic.mapper.substitutor import make_subst_func

    t_unit = _infer_lp_types(t_unit)
    knl = t_unit.default_entrypoint
    type_reader = TypeReader(knl, t_unit.callables_table)  # type: ignore[no-untyped-call]
    subst = knl.substitutions[rule_name]
    subst_expr = subst.expression
    # submap1: expand substitution invocations (expected by type inference mapper)
    submap1 = SubstitutionRuleExpander(knl.substitutions)
    # submap2: pass arguments (with appropriate dtypes) to the substitution of
    # interest
    submap2 = SubstitutionMapper(
        make_subst_func(
            {
                arg: np.iinfo(knl.index_dtype.numpy_dtype).max
                for arg in subst.arguments
            }
        )
    )

    subst_dtype = type_reader(submap2(submap1(subst_expr)))
    # type-ignore-reason: missing precise typing information in loopy
    return subst_dtype.numpy_dtype


class SubstitutionInvocationGetter(CombineMapper[frozenset[p.Call], []]):
    """
    Mapper to collect all the substitution invocations in an expression.

    .. attribute:: argument_substs

        The substitutions that are to be treated as arguments.
    """

    def __init__(self, argument_substs: frozenset[str]):
        self.argument_substs = argument_substs
        super().__init__()

    def combine(self, values: Iterable[frozenset[p.Call]]) -> frozenset[p.Call]:
        from functools import reduce

        return reduce(frozenset.union, values, frozenset())

    def map_call(self, expr: p.Call) -> frozenset[p.Call]:
        if cast("p.Variable", expr.function).name in self.argument_substs:
            return frozenset([expr])
        else:
            return super().map_call(expr)

    def map_algebraic_leaf(self, expr: Any) -> frozenset[p.Call]:
        return frozenset()

    def map_constant(self, expr: Any) -> frozenset[p.Call]:
        return frozenset()


def get_a_matched_einsum(
    t_unit: lp.TranslationUnit,
    kernel_name: str | None = None,
    insn_match: Any = None,
    argument_substitutions: frozenset[str] | None = None,
    long_dim_length: int = 500,
) -> tuple[BatchedEinsum, frozenbidict[str, str]]:
    """
    Returns a tuple of the form ``(matched_einsum, subst_map)`` where,
    ``matched_einsum`` is the batched einsum having a memory access pattern similar
    to the instructions in *insn_match* of *t_unit*, and, ``subst_map`` is a mapping
    from the variables in *t_unit* to the entities (i.e. indices, arguments) of
    *match_einsum*.

    :param t_unit: The subject translation unit which is being matched against
        *ref_einsum*.
    :param insn_match: A match expression as understood by
        :func:`loopy.match.parse_match` representing the subset of the
        kernel's instructions raised to a batched einsum.
    :param argument_substitutions: A :class:`frozenset` of substitutions that are to
        be matched as an Einstein-summation's argument. If *None*, any substitution
        rule invocation is treated as accessing an Einstein Summation's argument.
    :param long_dim_length: Axis length above which can be assumed to be long
        enough to be invariant of having any noticeable effects on the kernel's
        performance.

    .. note::

         See xxx (TODO) for a description of a translation unit with a einsum-like
         memory access pattern.
    """
    if kernel_name is None:
        if len(t_unit.entrypoints) != 1:
            raise EinsumTunitMatchError(
                "Must provide `kernel_name`, when"
                " the translation unit has more than"
                " 1 entrypoints."
            )
        else:
            kernel_name = t_unit.default_entrypoint.name

    assert isinstance(kernel_name, str)
    kernel = t_unit[kernel_name]

    argument_substitutions = argument_substitutions or frozenset(
        kernel.substitutions
    )
    assert isinstance(argument_substitutions, frozenset)

    from loopy.match import parse_match

    insn_match = parse_match(insn_match)
    insns = [insn for insn in kernel.instructions if insn_match(kernel, insn)]

    if len({insn.within_inames for insn in insns}) != 1:
        raise EinsumTunitMatchError(
            "Instructions forming the subject have"
            " more than 1 enclosing loop nest -- not"
            " allowed."
        )

    free_indices_set = {
        _get_indices_from_assignee(insn.assignees[0]) for insn in insns
    }
    if len(free_indices_set) != 1:
        raise EinsumTunitMatchError(
            "Instructions have differing free indices."
            " This is not allowed as the expressions"
            " are not batched-einsums-like."
        )

    (free_indices,) = free_indices_set

    if frozenset(free_indices) != insns[0].within_inames:
        raise EinsumTunitMatchError(
            "Einsum nested within an outer loop."
            " Such manipulation of batched einsums"
            " is not supported by feinsum."
        )

    get_reductions = ReductionCollector()
    batched_iname_set_to_operands: list[dict[tuple[str, ...], frozenset[str]]] = []
    subst_invokes_getter = SubstitutionInvocationGetter(argument_substitutions)

    # {{{ populate batched_access_to_operands

    for insn in insns:
        access_to_operands: dict[tuple[str, ...], set[str]] = {}
        redns_in_expr = get_reductions((insn.expression, tuple(insn.predicates)))
        if len(redns_in_expr) == 0:
            inner_expr = insn.expression
            ensm_inames = insn.within_inames
            redn_inames: frozenset[str] = frozenset()
        elif len(redns_in_expr) == 1:
            (redn_in_expr,) = redns_in_expr
            inner_expr = redn_in_expr.expr
            redn_inames = frozenset(redn_in_expr.inames)
            ensm_inames = insn.within_inames | redn_inames
        else:
            raise EinsumTunitMatchError(
                "More than one reductions found"
                " -> not within the grammar of expressions"
                " that can be matched by feinsum."
            )

        if isinstance(inner_expr, p.Product):
            from pymbolic.primitives import flattened_product

            flat_prod = flattened_product(inner_expr.children)
            if isinstance(flat_prod, p.Product):
                einsum_terms: tuple[p.Expression, ...] = flat_prod.children
            else:
                einsum_terms = (flat_prod,)
        else:
            einsum_terms = (inner_expr,)

        for term in einsum_terms:
            subst_invokes = subst_invokes_getter(term)
            if not subst_invokes:
                continue
            assert all(
                isinstance(subst_invoke, p.Call) for subst_invoke in subst_invokes
            )
            if not all(
                all(
                    isinstance(arg, p.Variable) and arg.name in ensm_inames
                    for arg in subst_invoke.parameters
                )
                for subst_invoke in subst_invokes
            ):
                raise EinsumTunitMatchError(
                    "Invocation to a substitution in "
                    f" '{term}' not called with the einstein"
                    " summation's indices -> not a part of"
                    " feinsum's grammar."
                )

            access_index_tuple_var: tuple[p.Variable, ...] = next(
                iter(subst_invokes)
            ).parameters  # type: ignore[assignment]

            if any(
                subst_invoke.parameters != access_index_tuple_var
                for subst_invoke in subst_invokes
            ):
                raise EinsumTunitMatchError(
                    "Not all substitution invocations in"
                    f" '{term}' are called with the same"
                    " arguments -> violates the feinsum's ."
                    " grammar."
                )

            access_index_tuple: tuple[str, ...] = tuple(
                idx.name for idx in access_index_tuple_var
            )
            subst_names = {
                cast("p.Variable", subst_invoke.function).name
                for subst_invoke in subst_invokes
            }
            access_to_operands.setdefault(access_index_tuple, set()).update(
                subst_names
            )

        batched_iname_set_to_operands.append(
            {k: frozenset(v) for k, v in access_to_operands.items()}
        )
    for iname_set_to_operands in batched_iname_set_to_operands:
        if len(iname_set_to_operands) != len(batched_iname_set_to_operands[0]):
            raise EinsumTunitMatchError(
                "Different einsums being performed by the instructions being"
                " matched against."
            )
        for iname_set, operands in iname_set_to_operands.items():
            if len(operands) != len(batched_iname_set_to_operands[0][iname_set]):
                raise EinsumTunitMatchError(
                    "Different einsums being performed by the instructions being"
                    " matched against."
                )

    # Step 1. Get a mapping from einsum inames to indices.
    ensm_inames = insns[0].within_inames | insn.reduction_inames()
    einsum_indices_generator = (chr(i) for i in range(ord("a"), ord("z") + 1))
    ensm_iname_to_index: dict[str, str] = {
        iname: next(einsum_indices_generator) for iname in ensm_inames
    }

    # Step 2. Get the $\mathcal{L}(i)$ for each $i$ in the new indices.
    iname_to_len = {
        iname: _get_iname_length(kernel, iname, long_dim_length, ensm_iname_to_index)
        for iname in ensm_inames
    }

    # Step 3. Combine these accesses to get the einsum expression.

    iname_sets: list[tuple[str, ...]] = []
    expanded_iname_sets: list[tuple[str, ...]] = []
    for iname_set, operands in sorted(
        batched_iname_set_to_operands[0].items(), key=lambda kxv: kxv[0]
    ):
        expanded_iname_sets.extend([iname_set] * len(operands))
        iname_sets.append(iname_set)
    einsum_subscripts = (
        ",".join(
            [
                "".join(ensm_iname_to_index[iname] for iname in in_iname_set)
                for in_iname_set in expanded_iname_sets
            ]
        )
        + "->"
        + "".join(ensm_iname_to_index[iname] for iname in free_indices)
    )
    arg_names: list[list[str]] = []
    for iname_set_to_operands in batched_iname_set_to_operands:
        arg_row = []
        for iname_set in iname_sets:
            arg_row.extend(sorted(iname_set_to_operands[iname_set]))
        assert len(arg_row) == len(expanded_iname_sets)
        arg_names.append(arg_row)

    # Step 4. Get the numeric data type for the substitution
    arg_to_dtype: dict[str, np.dtype[Any]] = {}

    for arg_row in arg_names:
        for arg in arg_row:
            if arg not in arg_to_dtype:
                arg_to_dtype[arg] = _get_dtype_for_subst_argument(
                    t_unit, kernel_name, arg
                )

    # Step 5. Get arg_shapes from $L$'s.
    arg_to_shape: dict[str, ShapeT] = {}
    for arg_row in arg_names:
        for arg, in_iname_set in zip(arg_row, expanded_iname_sets, strict=True):
            arg_shape = tuple(iname_to_len[iname] for iname in in_iname_set)
            if arg_to_shape.setdefault(arg, arg_shape) != arg_shape:
                raise EinsumTunitMatchError(
                    f"Substituion {arg} has multiple access maps in the instructions"
                    " being matched. This is not allowed."
                )

    # Step 6. Construct the batched einsum.
    from feinsum.make_einsum import array, batched_einsum

    beinsum = batched_einsum(
        einsum_subscripts,
        [
            [array(arg, arg_to_shape[arg], arg_to_dtype[arg]) for arg in arg_row]
            for arg_row in arg_names
        ],
    )

    # FIXME: Verify that the kernel's domain is indeed a dense-hypercube.
    output_names = ["_fe_out"] + [f"_fe_out_{i}" for i in range(len(insns) - 1)]

    subst_map = frozenbidict(
        {
            **ensm_iname_to_index,
            **{val: val for val in arg_to_dtype.keys()},
            **{
                insn.assignee_var_names()[0]: out_name
                for insn, out_name in szip(insns, output_names)
            },
        }
    )
    return beinsum, subst_map


def match_t_unit_to_einsum(
    t_unit: lp.TranslationUnit,
    einsum: BatchedEinsum,
    *,
    kernel_name: str | None = None,
    insn_match: Any = None,
    argument_substitutions: frozenset[str] | None = None,
    long_dim_length: int = 500,
) -> Mapping[str, str]:
    """
    Returns a mapping from the entities of *einsum* to the variables of the
    corresponding matched einsum in *t_unit*. See :func:`get_a_matched_einsum` for a
    subset of grammar of :mod:`loopy` kernels to be a matched as a batched einsum.
    """
    matched_einsum, var_in_tunit_to_var_in_matched_ensm = get_a_matched_einsum(
        t_unit, kernel_name, insn_match, argument_substitutions, long_dim_length
    )

    from feinsum.canonicalization import (
        get_substitution_mapping_between_isomorphic_batched_einsums,
    )

    isomorph_subst = get_substitution_mapping_between_isomorphic_batched_einsums(
        einsum, matched_einsum
    )
    matched_einsum_size_params = frozenset(
        {param.name for param in matched_einsum.all_size_params}
    )

    return Map(
        {
            var_in_ensm: var_in_tunit_to_var_in_matched_ensm.inv[var_in_matched_ensm]
            for var_in_ensm, var_in_matched_ensm in isomorph_subst.items()
            if var_in_matched_ensm not in matched_einsum_size_params
        }
    )


# }}}


# {{{ get_call_ids


class CallCollector(CombineMapper[frozenset[str], []]):
    def combine(self, values: Iterable[frozenset[str]]) -> frozenset[str]:
        from functools import reduce

        return reduce(frozenset.union, values, frozenset())

    def map_call(self, expr: p.Call) -> frozenset[str]:
        if isinstance(expr.function, p.Variable):
            return frozenset([expr.function.name]) | super().map_call(expr)
        else:
            return super().map_call(expr)

    def map_constant(self, expr: Any) -> frozenset[str]:
        return frozenset()

    def map_algebraic_leaf(self, expr: Any) -> frozenset[str]:
        return frozenset()


def get_call_ids(expr: p.Expression) -> frozenset[str]:
    """
    Returns the identifiers of the invoked functions in *expr*.
    """
    return CallCollector()(expr)


# }}}


# {{{ partition (copied from more-itertools)

Tpart = TypeVar("Tpart")


def partition(
    pred: Callable[[Tpart], bool], iterable: Iterable[Tpart]
) -> tuple[list[Tpart], list[Tpart]]:
    """
    Use a predicate to partition entries into false entries and true
    entries
    """
    # Inspired from https://docs.python.org/3/library/itertools.html
    # partition(is_odd, range(10)) --> 0 2 4 6 8   and  1 3 5 7 9
    from itertools import filterfalse, tee

    t1, t2 = tee(iterable)
    return list(filterfalse(pred, t1)), list(filter(pred, t2))


# }}}


# {{{ hoist_reduction_invariant_terms


class EinsumTermsHoister(IdentityMapper[[]]):
    """
    Mapper to hoist products out of a sum-reduction.

    .. attribute:: reduction_inames

        Inames of the reduction expressions to perform the hoisting.
    """

    def __init__(self, reduction_inames: frozenset[str]):
        super().__init__()
        self.reduction_inames = reduction_inames

    def map_reduction(self, expr: Reduction) -> p.Expression:
        if frozenset(expr.inames) != self.reduction_inames:
            return super().map_reduction(expr)

        from loopy.library.reduction import SumReductionOperation
        from loopy.symbolic import get_dependencies

        if isinstance(expr.operation, SumReductionOperation):
            rec_inner = self.rec(expr.expr)
            if isinstance(rec_inner, p.Product):
                from pymbolic.primitives import flattened_product

                flattened = flattened_product(rec_inner.children)
                if isinstance(flattened, p.Product):
                    multiplicative_terms: tuple[p.Expression, ...] = (
                        flattened.children
                    )
                else:
                    multiplicative_terms = (flattened,)

            else:
                multiplicative_terms = (rec_inner,)

            invariants, variants = partition(
                lambda x: bool(get_dependencies(x) & self.reduction_inames),
                multiplicative_terms,
            )
            if not variants:
                # -> everything is invariant
                assert p.is_arithmetic_expression(rec_inner)
                return rec_inner * Reduction(
                    expr.operation,
                    inames=tuple(expr.inames),
                    expr=1,  # FIXME: invalid dtype (not sure how?)
                    allow_simultaneous=expr.allow_simultaneous,
                )
            if not invariants:
                # -> nothing to hoist
                return Reduction(
                    expr.operation,
                    inames=tuple(expr.inames),
                    expr=self.rec(expr.expr),
                    allow_simultaneous=expr.allow_simultaneous,
                )

            return p.Product(tuple(invariants)) * Reduction(
                expr.operation,
                inames=tuple(expr.inames),
                expr=p.Product(tuple(variants)),
                allow_simultaneous=expr.allow_simultaneous,
            )
        else:
            return super().map_reduction(expr)


def hoist_invariant_multiplicative_terms_in_sum_reduction(
    kernel: LoopKernel, reduction_inames: str | frozenset[str], within: Any = None
) -> LoopKernel:
    """
    Hoists loop-invariant multiplicative terms in a sum-reduction expression.

    :arg reduction_inames: The inames over which reduction is performed that defines
        the reduction expression that is to be transformed.
    :arg within: A match expression understood by :func:`loopy.match.parse_match`
        that specifies the instructions over which the transformation is to be
        performed.
    """
    from loopy.transform.instruction import map_instructions

    if isinstance(reduction_inames, str):
        reduction_inames = frozenset([reduction_inames])

    if not (reduction_inames <= kernel.all_inames()):
        raise ValueError(
            f"Some inames in '{reduction_inames}' not a part of" " the kernel."
        )

    term_hoister = EinsumTermsHoister(reduction_inames)

    kernel = map_instructions(  # type: ignore[no-untyped-call]
        kernel,
        insn_match=within,
        f=lambda x: x.with_transformed_expressions(term_hoister),
    )
    return kernel


# }}}


# {{{ extract_multiplicative_terms_in_sum_reduction_as_subst


class ContainsSumReduction(CombineMapper[bool, []]):
    """
    Returns *True* only if the mapper maps over an expression containing a
    SumReduction operation.
    """

    def combine(self, values: Iterable[bool]) -> bool:
        return any(values)

    def map_reduction(self, expr: Reduction) -> bool:
        from loopy.library.reduction import SumReductionOperation

        return isinstance(expr.operation, SumReductionOperation) or self.rec(
            expr.expr
        )

    def map_variable(self, expr: p.Variable) -> bool:
        return False

    def map_algebraic_leaf(self, expr: Any) -> bool:
        return False


class MultiplicativeTermReplacer(IdentityMapper[[]]):
    """
    Primary mapper of
    :func:`extract_multiplicative_terms_in_sum_reduction_as_subst`.
    """

    def __init__(
        self,
        *,
        terms_filter: Callable[[p.Expression], bool],
        subst_name: str,
        subst_arguments: tuple[p.Expression, ...],
    ) -> None:
        self.subst_name = subst_name
        self.subst_arguments = subst_arguments
        self.terms_filter = terms_filter
        super().__init__()

        # mutable state to record the expression collected by the terms_filter
        self.collected_subst_rule: SubstitutionRule | None = None

    def map_reduction(self, expr: Reduction) -> p.Expression:
        from loopy.library.reduction import SumReductionOperation
        from loopy.symbolic import SubstitutionMapper

        if isinstance(expr.operation, SumReductionOperation):
            if self.collected_subst_rule is not None:
                # => there was already a sum-reduction operation -> raise
                raise ValueError(
                    "Multiple sum reduction expressions found -> not" " allowed."
                )

            if isinstance(expr.expr, p.Product):
                from pymbolic.primitives import flattened_product

                flattened = flattened_product(expr.expr.children)
                if isinstance(flattened, p.Product):
                    terms: tuple[p.Expression, ...] = flattened.children
                else:
                    terms = (flattened,)
            else:
                terms = (expr.expr,)

            unfiltered_terms, filtered_terms = partition(self.terms_filter, terms)
            submap = SubstitutionMapper(
                {
                    argument_expr: p.Variable(f"arg{i}")
                    for i, argument_expr in enumerate(self.subst_arguments)
                }.get
            )
            self.collected_subst_rule = SubstitutionRule(
                name=self.subst_name,
                arguments=tuple(f"arg{i}" for i in range(len(self.subst_arguments))),
                expression=submap(
                    p.Product(tuple(filtered_terms)) if filtered_terms else 1
                ),
            )
            return Reduction(
                expr.operation,
                tuple(expr.inames),
                p.Product(
                    (
                        p.Variable(self.subst_name)(*self.subst_arguments),
                        *unfiltered_terms,
                    )
                ),
                expr.allow_simultaneous,
            )
        else:
            return super().map_reduction(expr)


def extract_multiplicative_terms_in_sum_reduction_as_subst(
    kernel: LoopKernel,
    within: Any,
    subst_name: str,
    arguments: Sequence[p.Expression],
    terms_filter: Callable[[p.Expression], bool],
) -> LoopKernel:
    """
    Returns a copy of *kernel* with a new substitution named *subst_name* and
    *arguments* as arguments for the aggregated multiplicative terms in a
    sum-reduction expression.

    :arg within: A match expression understood by :func:`loopy.match.parse_match`
        to specify the instructions over which the transformation is to be
        performed.
    :arg terms_filter: A callable to filter which terms of the sum-reduction
        comprise the body of substitution rule.
    :arg arguments: The sub-expressions of the product of the filtered terms that
        form the arguments of the extract substitution rule in the same order.

    .. note::

        A ``LoopyError`` is raised if none or more than 1 sum-reduction expression
        appear in *within*.
    """
    from loopy.match import parse_match

    within = parse_match(within)

    matched_insns = [
        insn
        for insn in kernel.instructions
        if within(kernel, insn)
        and ContainsSumReduction()((insn.expression, tuple(insn.predicates)))
    ]

    if len(matched_insns) == 0:
        raise LoopyError(
            f"No instructions found matching '{within}'"
            " with sum-reductions found."
        )
    if len(matched_insns) > 1:
        raise LoopyError(
            f"More than one instruction found matching '{within}'"
            " with sum-reductions found -> not allowed."
        )

    (insn,) = matched_insns
    replacer = MultiplicativeTermReplacer(
        subst_name=subst_name,
        subst_arguments=tuple(arguments),
        terms_filter=terms_filter,
    )
    new_insn = insn.with_transformed_expressions(replacer)
    new_rule = replacer.collected_subst_rule
    new_substitutions = dict(kernel.substitutions).copy()
    if subst_name in new_substitutions:
        raise LoopyError(
            f"Kernel '{kernel.name}' already contains a substitution"
            f" rule named '{subst_name}'."
        )
    assert new_rule is not None
    new_substitutions[subst_name] = new_rule

    return kernel.copy(
        instructions=[
            new_insn if insn.id == new_insn.id else insn
            for insn in kernel.instructions
        ],
        substitutions=new_substitutions,
    )


# }}}

# {{{ decouple_domain


@for_each_kernel
def decouple_domain(
    kernel: LoopKernel, inames: Collection[str], parent_inames: Collection[str]
) -> LoopKernel:
    r"""
    Returns a copy of *kernel* with altered domains. The home domain of
    *inames* i.e. :math:`\mathcal{D}^{\text{home}}({\text{inames}})` is
    replaced with two domains :math:`\mathcal{D}_1` and :math:`\mathcal{D}_2`.
    :math:`\mathcal{D}_1` is the domain with dimensions corresponding to *inames*
    projected out and :math:`\mathcal{D}_2` is the domain with all the dimensions
    other than the ones corresponding to *inames* projected out.

    :arg inames: The inamaes to be decouple from their home domain.
    :arg parent_inames: Inames in :math:`\mathcal{D}^{\text{home}}({\text{inames}})`
        that will be used as additional parametric dimensions during the
        construction of :math:`\mathcal{D}_1`.

    .. note::

        - An error is raised if all the *inames* do not correspond to the same home
          domain of *kernel*.
        - It is the caller's responsibility to ensure that :math:`\mathcal{D}_1
          \cup \mathcal{D}_2 = \mathcal{D}^{\text{home}}({\text{inames}})`. If this
          criterion is violated this transformation would violate dependencies.
    """

    # {{{ sanity checks

    if not inames:
        raise LoopyError(
            "No inames were provided to decouple into" " a different domain."
        )
    if frozenset(parent_inames) & frozenset(inames):
        raise LoopyError("Inames cannot be appear in `inames` and `parent_inames`.")

    # }}}

    hdi = kernel.get_home_domain_index(next(iter(inames)))
    for iname in inames:
        if kernel.get_home_domain_index(iname) != hdi:
            raise LoopyError("inames are not a part of the same home domain.")

    all_dims = frozenset(kernel.domains[hdi].get_var_dict())
    for parent_iname in parent_inames:
        if parent_iname not in all_dims:
            raise LoopyError(
                f"Parent iname '{parent_iname}' not a part of the"
                f" corresponding home domain '{kernel.domains[hdi]}'."
            )

    dom1 = kernel.domains[hdi]
    dom2 = kernel.domains[hdi]

    for iname in sorted(all_dims):
        if iname in inames:
            dt, pos = dom1.get_var_dict()[iname]
            dom1 = dom1.project_out(dt, pos, 1)
        elif iname in parent_inames:
            dt, pos = dom2.get_var_dict()[iname]
            if dt != isl.dim_type.param:
                n_params = dom2.dim(isl.dim_type.param)
                dom2 = dom2.move_dims(isl.dim_type.param, n_params, dt, pos, 1)
        else:
            dt, pos = dom2.get_var_dict()[iname]
            dom2 = dom2.project_out(dt, pos, 1)

    new_domains = list(kernel.domains)
    new_domains[hdi] = dom1
    new_domains.append(dom2)
    kernel = kernel.copy(domains=new_domains)
    return kernel


# }}}

# vim: foldmethod=marker



from feinsum.canonicalization import canonicalize_einsum
from feinsum.cl_utils import make_fake_cl_context
from feinsum.codegen.loopy import (
    generate_loopy,
    generate_loopy_with_opt_einsum_schedule,
)
from feinsum.diagnostics import InvalidParameterError
from feinsum.einsum import (
    Array,
    BatchedEinsum,
    EinsumAxisAccess,
    FreeAxis,
    SummationAxis,
    get_opt_einsum_contraction_schedule,
    get_trivial_contraction_schedule,
)
from feinsum.loopy_utils import (
    get_a_matched_einsum,
    get_call_ids,
    match_t_unit_to_einsum,
)
from feinsum.make_einsum import array, batched_einsum, einsum
from feinsum.measure import (
    get_roofline_flop_rate,
    measure_giga_op_rate,
    stringify_comparison_vs_roofline,
    timeit,
    validate_batched_einsum_transform,
)
from feinsum.sql_utils import DEFAULT_DB, query
from feinsum.tuning import autotune
from feinsum.utils import IndexNameGenerator

__all__ = (
    "DEFAULT_DB",
    "Array",
    "BatchedEinsum",
    "EinsumAxisAccess",
    "FreeAxis",
    "IndexNameGenerator",
    "InvalidParameterError",
    "SummationAxis",
    "array",
    "autotune",
    "batched_einsum",
    "canonicalize_einsum",
    "einsum",
    "generate_loopy",
    "generate_loopy_with_opt_einsum_schedule",
    "get_a_matched_einsum",
    "get_call_ids",
    "get_opt_einsum_contraction_schedule",
    "get_roofline_flop_rate",
    "get_trivial_contraction_schedule",
    "make_fake_cl_context",
    "match_t_unit_to_einsum",
    "measure_giga_op_rate",
    "query",
    "stringify_comparison_vs_roofline",
    "timeit",
    "validate_batched_einsum_transform",
)



from typing import Any

import loopy as lp
import numpy as np

import feinsum as fnsm
from feinsum.tuning import IntParameter


@fnsm.tuning.einsum_arg("j_len", lambda e: e.args[0][0].shape[1])
@fnsm.tuning.transform_param("l_0_size", lambda e: IntParameter(32, 32))
def transform(
    t_unit: lp.TranslationUnit,
    j_len: int,
    l_0_size: int,
    insn_match: Any | None = None,
    kernel_name: str | None = None,
) -> lp.TranslationUnit:
    if l_0_size > 500:
        raise fnsm.InvalidParameterError(
            "Work-group size constraints for the" " hardware."
        )
    ref_einsum = fnsm.einsum("ij->i", fnsm.array("A", ("Ni", j_len), np.float64))
    subst_map = fnsm.match_t_unit_to_einsum(
        t_unit, ref_einsum, kernel_name=kernel_name, insn_match=insn_match
    )
    i = subst_map["i"]
    t_unit = lp.split_iname(t_unit, i, l_0_size, inner_tag="l.0", outer_tag="g.0")

    return t_unit


if __name__ == "__main__":
    import os

    import pyopencl as cl

    n_j = 4
    cl_ctx = cl.create_some_context()

    expr = fnsm.einsum("ij->i", fnsm.array("A", ("Ni", n_j), "float64"))

    fnsm.autotune(expr, os.path.abspath(__file__), cl_ctx, long_dim_length=200_000)



import logging
import math
from typing import Any

import loopy as lp

import feinsum as fnsm
import feinsum.loopy_utils as lp_utils
from feinsum.tuning import IntParameter

logger = logging.getLogger(__name__)


@fnsm.tuning.einsum_arg("ndim", lambda e: e.args[0][0].shape)
@fnsm.tuning.einsum_arg("ndof", lambda e: e.shape[1])
@fnsm.tuning.transform_param("n_e_per_wg", lambda e: IntParameter(2, 32))
@fnsm.tuning.transform_param(
    "nwork_items_per_e", lambda e: IntParameter(1, e.shape[1])
)
@fnsm.tuning.transform_param(
    "j_tiles", lambda e: IntParameter(1, math.ceil(e.shape[1] / 2))
)
@fnsm.tuning.transform_param(
    "i_tiles", lambda e: IntParameter(1, math.ceil(e.shape[1] / 2))
)
def transform(
    t_unit: lp.TranslationUnit,
    ndim: int,
    ndof: int,
    n_e_per_wg: int,
    nwork_items_per_e: int,
    i_tiles: int,
    j_tiles: int,
    insn_match: Any | None = None,
    kernel_name: str | None = None,
) -> lp.TranslationUnit:
    from loopy.match import parse_match

    kernel_name = kernel_name or t_unit.default_entrypoint.name

    within = parse_match(insn_match)
    knl = t_unit[kernel_name]
    (insn_id,) = [insn.id for insn in knl.instructions if within(knl, insn)]
    del knl

    ref_einsum = fnsm.einsum(
        "xre,rij,xej->ei",
        fnsm.array("J", (ndim, ndim, "Nel"), "float64"),
        fnsm.array("D", (ndim, ndof, ndof), "float64"),
        fnsm.array("u", (ndim, "Nel", ndof), "float64"),
    )

    # {{{ get corresponding variables in t_unit

    vng = t_unit.default_entrypoint.get_var_name_generator()
    ing = t_unit.default_entrypoint.get_instruction_id_generator()
    subst_map = fnsm.match_t_unit_to_einsum(
        t_unit, ref_einsum, insn_match=insn_match, kernel_name=kernel_name
    )
    i = subst_map["i"]
    j = subst_map["j"]
    e = subst_map["e"]
    D = subst_map["D"]
    u = subst_map["u"]
    x = subst_map["x"]
    r = subst_map["r"]

    j_inner, j_tile = vng(f"{j}_inner"), vng(f"{j}_tile")
    e_inner, e_outer = vng(f"{e}_inner"), vng(f"{e}_outer")
    i_inner, i_tile = vng(f"{i}_inner"), vng(f"{i}_tile")
    i_inner_inner, i_inner_outer = (vng(f"{i_inner}_inner"), vng(f"{i_inner}_outer"))
    rprftchD, iprftchD, jprftchD = (
        vng(f"{r}prftchD"),
        vng(f"{i}prftchD"),
        vng(f"{j}prftchD"),
    )
    D_fetch = vng(f"{D}_fetch")
    prcmpt_x_redn = ing(f"prcmpt_{x}_redn")
    e_prcmpt_subst, r_prcmpt_subst, j_prcmpt_subst = (
        vng(f"{e}prcmpt_subst"),
        vng(f"{r}prcmpt_subst"),
        vng(f"{j}prcmpt_subst"),
    )

    j_prcmpt_subst_inner, j_prcmpt_subst_outer = (
        vng(f"{j_prcmpt_subst}_inner"),
        vng(f"{j_prcmpt_subst}_outer"),
    )

    # }}}

    # {{{ term hoisting to match the flop count of opt_einsum

    from loopy.symbolic import get_dependencies
    from pymbolic import variables

    knl = t_unit[kernel_name]

    knl = lp.split_reduction_inward(knl, x)
    knl = lp_utils.hoist_invariant_multiplicative_terms_in_sum_reduction(knl, x)
    knl = lp_utils.extract_multiplicative_terms_in_sum_reduction_as_subst(
        knl,
        subst_name="subst",
        arguments=variables(f"{e} {j} {r}"),
        within=insn_match,
        terms_filter=lambda x: (
            (get_dependencies(x) & knl.all_inames()) <= {r, e, j}
        ),
    )

    knl = lp.split_iname(
        knl, j, math.ceil(ndof / j_tiles), outer_iname=j_tile, inner_iname=j_inner
    )

    t_unit = t_unit.with_kernel(knl)

    # }}}

    t_unit = lp.split_iname(
        t_unit,
        e,
        n_e_per_wg,
        inner_iname=e_inner,
        outer_iname=e_outer,
        inner_tag="l.1",
        outer_tag="g.0",
    )

    if 0:
        # TODO: This is valid only if len(i_inner_inner) is greater than
        # len(iprftch_D_inner). Enable this after fixing that.
        J = subst_map["J"]
        t_unit = lp.precompute(
            t_unit,
            J,
            sweep_inames=[x, r],
            precompute_outer_inames=frozenset({e_outer, e_inner, i_inner_inner}),
            temporary_address_space=lp.AddressSpace.PRIVATE,
            default_tag="unr",
            within=within,
        )

    # {{{ tile and prefetch D

    t_unit = lp.split_iname(
        t_unit,
        i,
        math.ceil(ndof / i_tiles),
        inner_iname=i_inner,
        outer_iname=i_tile,
        outer_tag="unr",
    )
    t_unit = lp.precompute(  # type: ignore[no-untyped-call]
        t_unit,
        D,
        [i_inner, r, j_inner],
        precompute_outer_inames=frozenset([e_outer, i_tile, j_tile]),
        precompute_inames=[rprftchD, iprftchD, jprftchD],
        temporary_address_space=lp.AddressSpace.LOCAL,
        temporary_name=D_fetch,
        within=within,
        default_tag=None,
    )
    t_unit = lp.split_iname(t_unit, jprftchD, n_e_per_wg, inner_tag="l.1")
    t_unit = lp.split_iname(t_unit, iprftchD, nwork_items_per_e, inner_tag="l.0")

    # }}}

    # {{{ precompute 'subst'

    t_unit = lp.precompute(  # type: ignore[no-untyped-call]
        t_unit,
        "subst",
        sweep_inames=[r, j_inner, e_inner],
        precompute_inames=[e_prcmpt_subst, j_prcmpt_subst, r_prcmpt_subst],
        precompute_outer_inames=frozenset({e_outer, j_tile}),
        default_tag=None,
        compute_insn_id=prcmpt_x_redn,
        temporary_address_space=lp.AddressSpace.LOCAL,
    )
    t_unit = lp.tag_inames(t_unit, {e_prcmpt_subst: "l.1"})

    # TODO: It might be worth exploring joining 'r_prcmpt_subst',
    # 'j_prcmpt_subst'.

    t_unit = lp.split_iname(
        t_unit,
        j_prcmpt_subst,
        nwork_items_per_e,
        inner_iname=j_prcmpt_subst_inner,
        outer_iname=j_prcmpt_subst_outer,
        inner_tag="l.0",
        outer_tag="unr",
    )

    # }}}

    t_unit = lp.split_iname(
        t_unit,
        i_inner,
        nwork_items_per_e,
        inner_iname=i_inner_inner,
        outer_iname=i_inner_outer,
        inner_tag="l.0",
        outer_tag="unr",
    )

    u_fetch = vng(f"{u}_fetch")
    t_unit = lp.precompute(  # type: ignore[no-untyped-call]
        t_unit,
        u,
        sweep_inames=[x, j_prcmpt_subst_outer],
        precompute_outer_inames=frozenset(
            [j_prcmpt_subst_inner, e_prcmpt_subst, e_outer, j_tile]
        ),
        precompute_inames=[vng("prftch_u_x"), None, vng("prftch_u_j")],
        temporary_address_space=lp.AddressSpace.PRIVATE,
        temporary_name=u_fetch,
        default_tag="unr",
    )

    # {{{ TODO: remove once github.com/inducer/loopy/issues/666 is resolved.

    t_unit = lp.realize_reduction(t_unit, insn_id_filter=insn_id)
    inames_to_duplicate = (
        frozenset({i_tile, i_inner_outer}) & t_unit[kernel_name].all_inames()
    )
    acc_name = f"acc_{r}_{j_tile}_{j_inner}"
    assert acc_name in t_unit[kernel_name].temporary_variables
    t_unit = lp.privatize_temporaries_with_inames(
        t_unit, inames_to_duplicate, only_var_names={acc_name}
    )

    t_unit = lp.duplicate_inames(
        t_unit,
        inames_to_duplicate,
        within=f"writes:{acc_name} and not reads:{acc_name}",
    )
    t_unit = lp.duplicate_inames(
        t_unit,
        inames_to_duplicate,
        within=f"reads:{acc_name} and not writes:{acc_name}",
    )

    # }}}

    t_unit = lp.tag_inames(t_unit, {r: "unr", x: "unr"})

    return t_unit


if __name__ == "__main__":
    import os
    from functools import partial

    import pyopencl as cl

    Ndim = 3
    Ndof = 35

    cl_ctx = cl.create_some_context()

    expr = fnsm.einsum(
        "xre,rij,xej->ei",
        fnsm.array("J", (Ndim, Ndim, "Nel"), "float64"),
        fnsm.array("D", (Ndim, Ndof, Ndof), "float64"),
        fnsm.array("u", (Ndim, "Nel", Ndof), "float64"),
    )

    if 1:
        fnsm.autotune(expr, os.path.abspath(__file__), cl_ctx)
    else:
        # Enable while debugging ->
        # evaluate a point in the parameter space.
        bound_transform = partial(
            transform,
            ndim=Ndim,
            ndof=Ndof,
            n_e_per_wg=27,
            nwork_items_per_e=9,
            i_tiles=4,
            j_tiles=4,
        )

        print(
            fnsm.stringify_comparison_vs_roofline(
                expr, transform=bound_transform, cl_ctx=cl_ctx
            )
        )

# vim: fdm=marker



"""
.. currentmodule:: feinsum.make_einsum

.. autofunction:: einsum
.. autofunction:: batched_einsum
"""

__copyright__ = """
Copyright (C) 2020 Andreas Kloeckner
Copyright (C) 2020 Matt Wala
Copyright (C) 2020 Xiaoyu Wei
Copyright (C) 2022 Kaushik Kulkarni
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import re
from collections.abc import Sequence
from typing import (
    Any,
)

import numpy as np
import numpy.typing as npt

from feinsum.einsum import (
    INT_CLASSES,
    Array,
    BatchedEinsum,
    ShapeComponentT,
    ShapeT,
    SizeParam,
)


def _preprocess_component(s: Any) -> ShapeComponentT:
    if isinstance(s, str):
        return SizeParam(s)
    elif isinstance(s, INT_CLASSES) and (s >= 0):
        return s
    elif isinstance(s, SizeParam):
        return s
    else:
        raise ValueError(f"Cannot infer shape component '{s}'.")


def _preprocess_shape(shape: Any) -> ShapeT:
    from collections.abc import Iterable

    if not isinstance(shape, Iterable):
        shape = (shape,)

    return tuple(_preprocess_component(d) for d in shape)


def array(name: str, shape: Any, dtype: npt.DTypeLike = "float64") -> Array:
    """
    Return an
    """
    return Array(name=name, shape=_preprocess_shape(shape), dtype=np.dtype(dtype))


EINSUM_FIRST_INDEX = re.compile(r"^\s*((?P<alpha>[a-zA-Z])|(?P<ellipsis>\.\.\.))\s*")


def _normalize_einsum_subscript(subscript: str, is_output: bool) -> tuple[str, ...]:
    """
    Normalizes the subscript of an einsum (provided in the explicit
    mode).
    """

    normalized_indices: list[str] = []
    acc = subscript.strip()
    while acc:
        match = EINSUM_FIRST_INDEX.match(acc)
        if match:
            if "alpha" in match.groupdict():
                normalized_indices.append(match.groupdict()["alpha"])
            else:
                assert "ellipsis" in match.groupdict()
                raise NotImplementedError("Broadcasting in einsums not supported")
            assert match.span()[0] == 0
            acc = acc[match.span()[-1] :]
        else:
            raise ValueError(
                f"Cannot parse '{acc}' in provided einsum" f" '{subscript}'."
            )

    if is_output:
        if len(set(normalized_indices)) != len(normalized_indices):
            raise ValueError(
                "Used an input more than once to refer to the"
                f" output axis in '{subscript}"
            )
    return tuple(normalized_indices)


def batched_einsum(
    subscripts: str,
    args: Sequence[Sequence[Array]],
) -> BatchedEinsum:
    """
    Returns a :class:`~feinsum.einsum.BatchedEinsum` with an interface similar to
    :func:`numpy.einsum`.

    :param subscripts: A :class:`str` describing the Einstein summation as
        accepted by :func:`numpy.einsum`.
    :param args: A b-long sequence each each comprising of n-long
        :class:`~feinsum.einsum.Array` instances, where, "b" is the number of
        einsums in the batched einsum, and, "n" is the number of array operands
        accepted by each of those b-einsums.
    """

    if "->" not in subscripts:
        # implicit-mode: output spec matched by alphabetical ordering of
        # indices (ewwwww)
        raise ValueError(
            "Missing -> in 'subscripts'. If the expected behavior"
            " is implicit mode, feinsum does not support it."
        )
    in_specs, out_spec = subscripts.split("->")
    out_idx_set = _normalize_einsum_subscript(out_spec, is_output=True)
    in_idx_sets = tuple(
        _normalize_einsum_subscript(in_spec, is_output=False)
        for in_spec in in_specs.split(",")
    )
    try:
        return BatchedEinsum(
            out_idx_set, in_idx_sets, tuple(tuple(arg_row) for arg_row in args)
        )
    except AssertionError as exc:
        raise TypeError(f"{exc}") from exc


def einsum(subscripts: str, *operands: Array) -> BatchedEinsum:
    """
    Returns a :class:`~feinsum.einsum.BatchedEinsum` with an interface similar to
    :func:`numpy.einsum`.
    """
    return batched_einsum(subscripts, [operands])


# vim: foldmethod=marker



"""
.. autofunction:: is_any_redn_dim_parametric
.. autofunction:: get_n_redn_dim

.. autoclass:: IndexNameGenerator
"""

import dataclasses as dc

from feinsum.einsum import BatchedEinsum, SizeParam, SummationAxis


def is_any_redn_dim_parametric(einsum: BatchedEinsum) -> bool:
    """
    .. testsetup::

        >>> from feinsum.utils import is_any_redn_dim_parametric
        >>> import numpy as np

    .. doctest::

        >>> import feinsum as f
        >>> ensm = f.einsum("ij,j->i",
        ...                 f.array((10, 4), "float64"),
        ...                 f.array(4, "float64"))
        >>> is_any_redn_dim_parametric(ensm)
        False
        >>> ensm = f.einsum("ij,jk->i",
        ...                 f.array((10, 4), "float64"),
        ...                 f.array((4, 20), "float64"))
        >>> is_any_redn_dim_parametric(ensm)
        False
        >>> ensm = f.einsum("i,j->i",
        ...                 f.array(125, "float64"),
        ...                 f.array(np.inf, "float64"))
        >>> is_any_redn_dim_parametric(ensm)
        True
    """
    return any(
        isinstance(dim_len, SizeParam)
        and isinstance(einsum.index_to_access_descr[idx], SummationAxis)
        for idx, dim_len in einsum.index_to_dim_length.items()
    )


def get_n_redn_dim(ensm: BatchedEinsum) -> int:
    """
    Returns the number of reduction indices in *ensm*.
    """
    return len(
        {
            idx
            for idx, access in ensm.index_to_access_descr.items()
            if isinstance(access, SummationAxis)
        }
    )


@dc.dataclass
class IndexNameGenerator:
    """
    Generates indices to be fed into an einstein summation.

    .. attribute:: banned_names

        A :class:`frozenset` of index names that will not be generated.

    .. doctest::

        >>> from feinsum import IndexNameGenerator
        >>> idx_name_gen = IndexNameGenerator(frozenset({'c'}))
        >>> idx_name_gen()
        'a'
        >>> idx_name_gen()
        'b'
        >>> idx_name_gen()
        'd'
    """

    banned_names: frozenset[str] = dc.field(default=frozenset())
    counter: int = dc.field(init=False, default=0)

    def __call__(self) -> str:
        if self.counter == 26:
            raise RuntimeError("All indices have been exhausted")

        new_name = chr(97 + self.counter)
        self.counter += 1

        if new_name in self.banned_names:
            return self()
        else:
            return new_name



from collections.abc import Mapping

# Mapping from device name to peak flop rate in GFlOps/s
DEV_TO_PEAK_GFLOPS: Mapping[str, Mapping[str, float]]
DEV_TO_PEAK_GFLOPS = {
    "NVIDIA TITAN V": {
        "float32": 12288,
        "float64": 6144,
    },
    "NVIDIA GeForce GTX 1650": {
        "float32": 3916.0,
        "float64": 122.4,
    },
}

# Mapping from device name to peak bandwidth in GB/s
DEV_TO_PEAK_BW = {
    "NVIDIA TITAN V": 652.8,
    "NVIDIA GeForce GTX 1650": 192.0,
}



from urllib.request import urlopen

# {{{ Copied from https://raw.githubusercontent.com/inducer/sphinxconfig/main/sphinxconfig.py

from os.path import basename as _basename, dirname as _dirname

# NOTE: these are only imported for type checking
from docutils.nodes import Text, TextElement
from sphinx.addnodes import pending_xref
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment


html_show_sourcelink = True

project = _basename(_dirname(_dirname(__file__)))

autoclass_content = "class"

copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

sphinxconfig_missing_reference_aliases: dict[str, str] = {}


def linkcode_resolve(
        domain: str,
        info: dict[str, str],
        linkcode_url: str | None = None,
        ) -> str | None:
    import inspect
    import os
    import sys

    if domain != "py" or not info["module"]:
        return None

    submodname = info["module"]
    topmodname = submodname.split(".")[0]
    fullname = info["fullname"]

    topmod = sys.modules.get(topmodname)
    submod = sys.modules.get(submodname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            obj = getattr(obj, part)
        except Exception:
            return None

    try:
        modpath = os.path.dirname(os.path.dirname(inspect.getsourcefile(topmod)))
        filepath = os.path.relpath(inspect.getsourcefile(obj), modpath)
    except Exception:
        return None

    if filepath is None:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except OSError:
        return None
    else:
        linestart, linestop = lineno, lineno + len(source) - 1

    if linkcode_url is None:
        linkcode_url = (
            f"https://github.com/kaushikcfd/{project}/blob/"
            + "main"
            + "/src/{filepath}#L{linestart}-L{linestop}"  # noqa: RUF027
        )

    return linkcode_url.format(
        filepath=filepath, linestart=linestart, linestop=linestop
    )


def process_autodoc_missing_reference(
        app: Sphinx,
        env: BuildEnvironment,
        node: pending_xref,
        contnode: Text | TextElement
        ) -> TextElement | None:
    """Fix missing references due to string annotations.

    Users of this function should add it to the ``missing-reference`` event as

    .. code:: python

        def setup(app: Sphinx) -> None:
            app.connect("missing-reference", process_autodoc_missing_reference)

    This uses an alias dictionary called ``sphinxconfig_missing_reference_aliases``
    that maps each unknown type to an reference type and actual type full name.
    For example, say that the ``numpy.float64`` reference is not recognized
    properly. We know that this object is in the ``numpy`` module as an attribute.
    Then, we write:

    .. code:: python

        autodoc_missing_reference_aliases: dict[str, str] = {
            "numpy.float64": "py:attr:numpy.float64",
        }
    """
    from sphinx.util import logging
    logger = logging.getLogger("sphinxconfig")

    # check if this is a known alias
    target = node["reftarget"]
    if target not in sphinxconfig_missing_reference_aliases:
        return None

    # parse alias
    fullname = sphinxconfig_missing_reference_aliases[target]
    parts = fullname.split(":")
    if len(parts) == 1:
        domain = "py"
        reftype = "obj"
        reftarget, = parts
    elif len(parts) == 2:
        domain = "py"
        reftype, reftarget = parts
    elif len(parts) == 3:
        domain, reftype, reftarget = parts
    else:
        logger.error("Could not parse alias for '%s': '%s'.", target, fullname)
        return None

    if domain != "py":
        logger.error("Domain not supported for '%s': '%s'.", target, domain)
        return None

    # parse module and object from reftarget
    parts = reftarget.split(".")
    if len(parts) == 1:
        inventory = module = objname = parts[0]
    elif len(parts) >= 2:
        inventory = parts[0]
        module = ".".join(parts[:-1])
        objname = parts[-1]
    else:
        logger.error("Could not parse reftarget for '%s': '%s'.", target, reftarget)
        return None

    # rewrite attributes
    node.attributes["refdomain"] = domain
    node.attributes["reftype"] = reftype
    node.attributes[f"{domain}:module"] = module

    if isinstance(contnode, Text):
        # NOTE: if the format is "short", we insist on using just the object name.
        # This also makes sure that the text matches the reftarget, in case the
        # object was renamed somehow (e.g. from `import Object as _Object`).
        if app.config.autodoc_typehints_format == "short":
            contnode = Text(objname)
        else:
            contnode = Text(reftarget)

    # resolve reference
    from sphinx.ext import intersphinx

    if intersphinx.inventory_exists(env, inventory):
        node.attributes["reftarget"] = reftarget

        new_node = intersphinx.resolve_reference_in_inventory(
            env, inventory, node, contnode)
    else:
        domain = env.get_domain(domain)
        new_node = domain.resolve_xref(
            env, node["refdoc"], app.builder, reftype, objname,
            node, contnode)

    return new_node


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.doctest",
    "sphinx.ext.mathjax",
    "sphinx_copybutton",
]

__all__ = (
    "autoclass_content",
    "copybutton_prompt_is_regexp",
    "copybutton_prompt_text",
    "extensions",
    "html_show_sourcelink",
    "html_theme",
    "linkcode_resolve",
    "project",
)

# }}}


html_theme = "sphinx_book_theme"
html_theme_options = {
}
html_static_path = ["static"]
html_logo = "static/logo.svg"
copyright = "2025, feinsum Contributors"
author = "feinsum Contributors"
release = "2025.3"

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pyopencl": ("https://documen.tician.de/pyopencl/", None),
    "pytools": ("https://documen.tician.de/pytools/", None),
    "pymbolic": ("https://documen.tician.de/pymbolic/", None),
    "loopy": ("https://documen.tician.de/loopy/", None),
    "islpy": ("https://documen.tician.de/islpy/", None),
    "opt_einsum": ("https://optimized-einsum.readthedocs.io/en/stable/", None),
}

nitpick_ignore_regex = [
    ["py:class", r"numpy.(u?)int[\d]+"],
    ["py:class", r"numpy.(_?)typing.(.+)"],
    # As of 2022-09-15, it doesn't look like there's sphinx documentation
    # available.
    ["py:class", r"immutables\.(.+)"],
    ["py:class", r"bidict\.(.+)"],
]

# vim: fdm=marker



import logging

import loopy as lp
import numpy as np
import pyopencl as cl

import feinsum as f
from feinsum.tuning import IntParameter

logging.basicConfig(level="INFO")


@f.tuning.transform_param("block_size_div_32", lambda ensm: IntParameter(1, 5))
def transform(t_unit, block_size_div_32, insn_match=None, kernel_name=None):
    ref_einsum = f.batched_einsum(
        "ij,j->i",
        [(np.inf, 4), (4,)],
        dtypes="float64",
        use_matrix=[
            [{"a"}, {"c"}],
            [{"b"}, {"c"}],
        ],
    )
    subst_map = f.match_t_unit_to_einsum(t_unit, ref_einsum)
    i = subst_map["i"]
    return lp.split_iname(
        t_unit, i, block_size_div_32 * 32, inner_tag="l.0", outer_tag="g.0"
    )


if __name__ == "__main__":
    import os

    cl_ctx = cl.create_some_context()

    expr = f.batched_einsum(
        "ij,j->i",
        [(np.inf, 4), (4,)],
        dtypes="float64",
        use_matrix=[
            [{"a"}, {"c"}],
            [{"b"}, {"c"}],
        ],
    )
    f.autotune(
        expr,
        os.path.abspath(__file__),
        cl_ctx,
        long_dim_length=500_000,
        stop_after=5,
    )



import logging
import math
from typing import TYPE_CHECKING, Any, cast

import islpy as isl
import loopy as lp
from more_itertools import zip_equal as szip

import feinsum as fnsm
import feinsum.loopy_utils as lp_utils
from feinsum.tuning import IntParameter

if TYPE_CHECKING:
    from collections.abc import Collection

logger = logging.getLogger(__name__)


def transform_with_single_j_tile_i_tile(
    t_unit: lp.TranslationUnit,
    nface: int,
    nvoldof: int,
    nfacedof: int,
    nfields: int,
    n_e_per_wg: int,
    nwork_items_per_e: int,
    n_stmt_tile: int,
    insn_match: Any | None = None,
    kernel_name: str | None = None,
) -> lp.TranslationUnit:
    import loopy.match as lp_match
    from loopy.symbolic import get_dependencies
    from more_itertools import chunked
    from pymbolic import variables

    kernel_name = kernel_name or t_unit.default_entrypoint.name

    within = lp_match.parse_match(insn_match)

    ref_einsum = fnsm.batched_einsum(
        "ifj,fe,fej->ei",
        [
            [
                fnsm.array("L", (nvoldof, nface, nfacedof)),
                fnsm.array("J", (nface, "Nel")),
                fnsm.array(f"v{i}", (nface, "Nel", nfacedof)),
            ]
            for i in range(nfields)
        ],
    )
    len_stmt_tile = math.ceil(nfields / n_stmt_tile)

    # {{{ get corresponding variables in t_unit

    vng = t_unit[kernel_name].get_var_name_generator()
    ing = t_unit[kernel_name].get_instruction_id_generator()
    subst_map = fnsm.match_t_unit_to_einsum(
        t_unit, ref_einsum, insn_match=insn_match, kernel_name=kernel_name
    )
    i = subst_map["i"]
    j = subst_map["j"]
    e = subst_map["e"]
    f = subst_map["f"]
    L = subst_map["L"]
    e_outer, e_inner = vng(f"{e}_outer"), vng(f"{e}_inner")
    fields = [subst_map[f"v{i}"] for i in range(nfields)]
    outputs = [subst_map["_fe_out"]] + [
        subst_map[f"_fe_out_{i}"] for i in range(nfields - 1)
    ]
    subst_names = {field: vng("subst_hoist") for field in fields}
    i_s = [vng(i) for _ in range(n_stmt_tile)]
    f_s = [vng(f) for _ in range(n_stmt_tile)]
    j_s = [vng(j) for _ in range(n_stmt_tile)]
    e_s = [vng(e) for _ in range(n_stmt_tile)]
    e_outer_names = [vng("e_outer") for _ in range(n_stmt_tile)]
    e_inner_names = [vng("e_inner") for _ in range(n_stmt_tile)]
    i_outer_names = [vng(f"{i}_outer") for _ in range(n_stmt_tile)]
    i_inner_names = [vng(f"{i}_inner") for _ in range(n_stmt_tile)]
    L_fetch = vng(f"{L}_fetch")
    # prefetch_L_insns_ids = [vng(f"prftch_{L}") for _ in range(n_stmt_tile)]
    i_stmt_to_subst_prcmpt_tmp = [vng("prcmpt_stage1") for _ in range(len_stmt_tile)]
    i_stmt_tile_to_e_prcmpt_stage1 = [
        vng(f"{e}_prcmpt_stage1") for _ in range(n_stmt_tile)
    ]
    i_stmt_tile_to_f_prcmpt_stage1 = [
        vng(f"{f}_prcmpt_stage1") for _ in range(n_stmt_tile)
    ]
    i_stmt_tile_to_j_prcmpt_stage1 = [
        vng(f"{j}_prcmpt_stage1") for _ in range(n_stmt_tile)
    ]
    compute_fxj_id = {field: ing(f"compute_fxj_{field}") for field in fields}

    # }}}

    # {{{ splitting fields across outer_statement_tiles

    i_stmt_tile_to_fields = [
        list(el) for el in chunked(fields, math.ceil(nfields / n_stmt_tile))
    ]
    i_stmt_tile_to_outputs = [
        list(el) for el in chunked(outputs, math.ceil(nfields / n_stmt_tile))
    ]
    assert all(len(el) <= len_stmt_tile for el in i_stmt_tile_to_fields)
    assert all(
        len(el1) == len(el2)
        for el1, el2 in szip(i_stmt_tile_to_fields, i_stmt_tile_to_outputs)
    )

    # }}}

    knl = t_unit[kernel_name]

    for fields_in_tile, outputs_in_tile in szip(
        i_stmt_tile_to_fields, i_stmt_tile_to_outputs
    ):
        for field, output in szip(fields_in_tile, outputs_in_tile):
            subst_name = subst_names[field]
            insn_match = lp_match.And((within, lp_match.Writes(output)))
            knl = lp_utils.extract_multiplicative_terms_in_sum_reduction_as_subst(
                knl,
                within=insn_match,
                subst_name=subst_name,
                arguments=variables(f"{f} {e} {j}"),
                terms_filter=lambda x: (get_dependencies(x) & knl.all_inames())
                <= {f, e, j},
            )
    t_unit = t_unit.with_kernel(knl)

    f_prftchL, i_prftchL, j_prftchL = (
        vng(f"{f}prftch{L}"),
        vng(f"{i}prftch{L}"),
        vng(f"{j}prftch{L}"),
    )

    t_unit = lp.precompute(  # type: ignore[no-untyped-call]
        t_unit,
        L,
        sweep_inames=[f, i, j],
        precompute_outer_inames=frozenset(),
        precompute_inames=[i_prftchL, f_prftchL, j_prftchL],
        default_tag=None,
        within=within,
        temporary_name=L_fetch,
    )

    t_unit = lp.split_iname(
        t_unit, i_prftchL, n_e_per_wg, inner_tag="l.1", outer_tag="unr"
    )
    t_unit = lp.split_iname(
        t_unit, j_prftchL, nwork_items_per_e, inner_tag="l.0", outer_tag="unr"
    )

    for i_stmt_tile, fields_in_tile in enumerate(i_stmt_tile_to_fields):
        new_j = j_s[i_stmt_tile]
        new_f = f_s[i_stmt_tile]
        new_i = i_s[i_stmt_tile]
        new_e = e_s[i_stmt_tile]
        e_inner = e_inner_names[i_stmt_tile]
        e_outer = e_outer_names[i_stmt_tile]
        i_inner_name = i_inner_names[i_stmt_tile]
        i_outer_name = i_outer_names[i_stmt_tile]

        outputs_insn_match = lp_match.And(
            (
                within,
                lp_match.Or(
                    tuple(
                        lp_match.Writes(output)
                        for output in i_stmt_tile_to_outputs[i_stmt_tile]
                    )
                ),
            )
        )
        t_unit = lp.duplicate_inames(
            t_unit,
            [e, f, i, j],
            within=outputs_insn_match,
            new_inames=[new_e, new_f, new_i, new_j],
        )
        parent_inames = cast(
            "Collection[str]",
            t_unit[kernel_name]
            .get_inames_domain(e)
            .get_var_names(isl.dim_type.param),
        )
        assert all(isinstance(parent_iname, str) for parent_iname in parent_inames)
        t_unit = t_unit.with_kernel(
            lp_utils.decouple_domain(
                t_unit[kernel_name],
                [new_e, new_f, new_i, new_j],
                parent_inames=parent_inames,
            ),
        )

        t_unit = lp.split_iname(
            t_unit,
            new_e,
            n_e_per_wg,
            inner_iname=e_inner,
            outer_iname=e_outer,
            inner_tag="l.1",
            outer_tag="g.0",
        )

        for istmt, field in enumerate(fields_in_tile):
            subst_name = subst_names[field]
            t_unit = lp.precompute(  # type: ignore[no-untyped-call]
                t_unit,
                subst_name,
                sweep_inames=[e_inner, new_j, new_f],
                temporary_address_space=lp.AddressSpace.LOCAL,
                temporary_name=i_stmt_to_subst_prcmpt_tmp[istmt],
                precompute_outer_inames=frozenset({e_outer}),
                precompute_inames=[
                    i_stmt_tile_to_f_prcmpt_stage1[i_stmt_tile],
                    i_stmt_tile_to_e_prcmpt_stage1[i_stmt_tile],
                    i_stmt_tile_to_j_prcmpt_stage1[i_stmt_tile],
                ],
                compute_insn_id=compute_fxj_id[field],
                default_tag=None,
            )

        t_unit = lp.tag_inames(
            t_unit, {i_stmt_tile_to_e_prcmpt_stage1[i_stmt_tile]: "l.1"}
        )
        t_unit = lp.split_iname(
            t_unit,
            i_stmt_tile_to_j_prcmpt_stage1[i_stmt_tile],
            nwork_items_per_e,
            inner_tag="l.0",
            # TODO: uncommenting this leads to 20% slow down in POCL & 10%
            # speedup in Nv-CL.
            # outer_tag="unr"
        )

        t_unit = lp.split_iname(
            t_unit,
            new_i,
            nwork_items_per_e,
            inner_iname=i_inner_name,
            outer_iname=i_outer_name,
            inner_tag="l.0",
            outer_tag="unr",
        )
        t_unit = lp.prioritize_loops(t_unit, [new_j, new_f])
        # t_unit = lp.tag_inames(t_unit, {new_f: "unr"})

    for i_stmt_tile in range(1, n_stmt_tile):
        predecessors = lp_match.And(
            (
                within,
                lp_match.Or(
                    tuple(
                        lp_match.Writes(output)
                        for output in i_stmt_tile_to_outputs[i_stmt_tile - 1]
                    )
                ),
            )
        )
        successors = lp_match.Or(
            tuple(
                lp_match.Id(compute_fxj_id[field])
                for field in i_stmt_tile_to_fields[i_stmt_tile]
            )
        )
        t_unit = lp.add_dependency(t_unit, successors, predecessors)

    t_unit = lp.add_inames_to_insn(
        t_unit, inames=e_outer_names[0], insn_match=lp_match.Writes(L_fetch)
    )

    return t_unit


@fnsm.tuning.einsum_arg("nface", lambda e: e.args[0][2].shape[0])
@fnsm.tuning.einsum_arg("nvoldof", lambda e: e.shape[1])
@fnsm.tuning.einsum_arg("nfacedof", lambda e: e.args[0][2].shape[2])
@fnsm.tuning.einsum_arg("nfields", lambda e: e.b)
@fnsm.tuning.transform_param("n_e_per_wg", lambda e: IntParameter(2, 32))
@fnsm.tuning.transform_param(
    "nwork_items_per_e", lambda e: IntParameter(1, e.args[0][2].shape[2])
)
@fnsm.tuning.transform_param(
    "n_stmt_tile", lambda e: IntParameter(1, math.ceil(e.b))
)
@fnsm.tuning.transform_param(
    "n_i_tile", lambda e: IntParameter(1, math.ceil(e.shape[1] / 2))
)
@fnsm.tuning.transform_param(
    "n_j_tile", lambda e: IntParameter(1, math.ceil(e.args[0][2].shape[2] / 2))
)
def transform(
    t_unit: lp.TranslationUnit,
    nface: int,
    nvoldof: int,
    nfacedof: int,
    nfields: int,
    n_e_per_wg: int,
    nwork_items_per_e: int,
    n_stmt_tile: int,
    n_i_tile: int,
    n_j_tile: int,
    insn_match: Any | None = None,
    kernel_name: str | None = None,
) -> lp.TranslationUnit:
    import loopy.match as lp_match
    from loopy.symbolic import get_dependencies
    from more_itertools import chunked
    from pymbolic import variables

    n_stmt_tile = math.ceil(nfields / math.ceil(nfields / n_stmt_tile))

    if n_j_tile == 1 and n_i_tile == 1:
        return transform_with_single_j_tile_i_tile(
            t_unit,
            nface,
            nvoldof,
            nfacedof,
            nfields,
            n_e_per_wg,
            nwork_items_per_e,
            n_stmt_tile,
            insn_match=insn_match,
            kernel_name=kernel_name,
        )
    kernel_name = kernel_name or t_unit.default_entrypoint.name

    within = lp_match.parse_match(insn_match)
    ref_einsum = fnsm.batched_einsum(
        "ifj,fe,fej->ei",
        [
            [
                fnsm.array("L", (nvoldof, nface, nfacedof)),
                fnsm.array("J", (nface, "Nel")),
                fnsm.array(f"v{i}", (nface, "Nel", nfacedof)),
            ]
            for i in range(nfields)
        ],
    )
    len_stmt_tile = math.ceil(nfields / n_stmt_tile)
    len_j_tile = math.ceil(nfacedof / n_j_tile)
    len_i_tile = math.ceil(nvoldof / n_i_tile)

    # {{{ get corresponding variables in t_unit

    vng = t_unit.default_entrypoint.get_var_name_generator()
    ing = t_unit.default_entrypoint.get_instruction_id_generator()
    subst_map = fnsm.match_t_unit_to_einsum(
        t_unit, ref_einsum, insn_match=insn_match, kernel_name=kernel_name
    )
    i = subst_map["i"]
    j = subst_map["j"]
    e = subst_map["e"]
    f = subst_map["f"]
    L = subst_map["L"]
    fields = [subst_map[f"v{i}"] for i in range(nfields)]
    outputs = [subst_map["_fe_out"]] + [
        subst_map[f"_fe_out_{i}"] for i in range(nfields - 1)
    ]
    subst_names = {field: vng("subst_hoist") for field in fields}
    e_s = [vng(e) for _ in range(n_stmt_tile)]
    i_s = [vng(i) for _ in range(n_stmt_tile)]
    j_s = [vng(j) for _ in range(n_stmt_tile)]
    f_s = [vng(f) for _ in range(n_stmt_tile)]
    j_tile_names = [vng(f"{j}_tile") for _ in range(n_stmt_tile)]
    j_inner_names = [vng(f"{j}_inner") for _ in range(n_stmt_tile)]
    i_tile_names = [vng(f"{i}_tile") for _ in range(n_stmt_tile)]
    i_inner_names = [vng(f"{i}_inner") for _ in range(n_stmt_tile)]
    i_inner_inner_names = [vng(f"{i}_inner_inner") for _ in range(n_stmt_tile)]
    i_inner_outer_names = [vng(f"{i}_inner_outer") for _ in range(n_stmt_tile)]
    L_fetch = vng(f"{L}_fetch")
    prefetch_L_insns_ids = [ing(f"prftch_{L}") for _ in range(n_stmt_tile)]
    e_outer_names = [vng(f"{e}_outer") for _ in range(n_stmt_tile)]
    e_inner_names = [vng(f"{e}_inner") for _ in range(n_stmt_tile)]
    i_stmt_to_subst_prcmp_tmp = [vng("prcmpt_stage1") for _ in range(len_stmt_tile)]
    i_stmt_tile_to_itile_init = [vng(f"{i}_tile_init") for _ in range(n_stmt_tile)]
    i_stmt_tile_to_i_inner_outer_assign = [
        vng(f"{i}_inner_outer_assign") for _ in range(n_stmt_tile)
    ]
    i_stmt_tile_to_itile_assign = [
        vng(f"{i}_tile_assign") for _ in range(n_stmt_tile)
    ]
    i_stmt_tile_to_i_inner_outer_init = [
        vng(f"{i}_inner_outer_init") for _ in range(n_stmt_tile)
    ]
    i_stmt_tile_to_e_prcmpt_stage1 = [
        vng(f"{e}_prcmpt_stage1") for _ in range(n_stmt_tile)
    ]
    i_stmt_tile_to_f_prcmpt_stage1 = [
        vng(f"{f}_prcmpt_stage1") for _ in range(n_stmt_tile)
    ]
    i_stmt_tile_to_j_prcmpt_stage1 = [
        vng(f"{j}_prcmpt_stage1") for _ in range(n_stmt_tile)
    ]

    # }}}

    # {{{ splitting fields across outer_statement_tiles

    i_stmt_tile_to_fields = [
        list(el) for el in chunked(fields, math.ceil(nfields / n_stmt_tile))
    ]
    i_stmt_tile_to_outputs = [
        list(el) for el in chunked(outputs, math.ceil(nfields / n_stmt_tile))
    ]
    assert all(len(el) <= len_stmt_tile for el in i_stmt_tile_to_fields)
    assert all(
        len(el1) == len(el2)
        for el1, el2 in szip(i_stmt_tile_to_fields, i_stmt_tile_to_outputs)
    )

    # }}}

    # {{{ split the kernel into disparate chunks

    for outputs_in_tile, new_i, new_f, new_e, new_j in szip(
        i_stmt_tile_to_outputs, i_s, f_s, e_s, j_s
    ):
        insn_match = lp_match.And(
            (
                within,
                lp_match.Or(
                    tuple(
                        lp_match.Writes(output_name)
                        for output_name in outputs_in_tile
                    )
                ),
            )
        )
        # FIXME: These should pull the duplicated inames into basic sets of their
        # own -> brings down the computational time.
        t_unit = lp.duplicate_inames(
            t_unit,
            (i, f, e, j),
            within=insn_match,
            new_inames=[new_i, new_f, new_e, new_j],
        )
        t_unit = t_unit.with_kernel(
            lp_utils.decouple_domain(
                t_unit[kernel_name],
                [new_e, new_i, new_j, new_f],
                parent_inames=cast(
                    "Collection[str]",
                    t_unit[kernel_name]
                    .get_inames_domain(e)
                    .get_var_names(isl.dim_type.param),
                ),
            )
        )

    t_unit = lp.remove_unused_inames(t_unit)

    # }}}

    for i_stmt_tile, (fields_in_tile, outputs_in_tile) in enumerate(
        szip(i_stmt_tile_to_fields, i_stmt_tile_to_outputs)
    ):
        new_i = i_s[i_stmt_tile]
        new_j = j_s[i_stmt_tile]
        new_f = f_s[i_stmt_tile]
        new_e = e_s[i_stmt_tile]
        j_tile_name = j_tile_names[i_stmt_tile]
        j_inner_name = j_inner_names[i_stmt_tile]
        i_tile_name = i_tile_names[i_stmt_tile]
        i_inner_name = i_inner_names[i_stmt_tile]
        i_inner_inner_name = i_inner_inner_names[i_stmt_tile]
        i_inner_outer_name = i_inner_outer_names[i_stmt_tile]
        f_prftchL, i_prftchL, j_prftchL = (
            vng(f"{f}prftch{L}"),
            vng(f"{i}prftch{L}"),
            vng(f"{j}prftch{L}"),
        )

        # There's a problem here. The accumulator names are sort of random
        # here which is obnoxious. We probably need to use some metadata here.

        knl = t_unit[kernel_name]

        for field, output in szip(fields_in_tile, outputs_in_tile):
            subst_name = subst_names[field]
            # FIXME: use precompute inames based on which inner statement tile
            # does the field belong to.

            insn_match = lp_match.And((within, lp_match.Writes(output)))

            knl = lp_utils.extract_multiplicative_terms_in_sum_reduction_as_subst(
                knl,
                within=insn_match,
                subst_name=subst_name,
                arguments=variables(f"{new_e} {new_j} {new_f}"),
                terms_filter=lambda x: (get_dependencies(x) & knl.all_inames())
                <= {new_e, new_f, new_j},
            )

        t_unit = t_unit.with_kernel(knl)

        t_unit = lp.split_iname(
            t_unit,
            new_j,
            len_j_tile,
            outer_iname=j_tile_name,
            inner_iname=j_inner_name,
        )
        t_unit = lp.split_iname(
            t_unit,
            new_i,
            len_i_tile,
            outer_iname=i_tile_name,
            inner_iname=i_inner_name,
        )
        t_unit = lp.split_iname(
            t_unit,
            new_e,
            n_e_per_wg,
            inner_iname=e_inner_names[i_stmt_tile],
            outer_iname=e_outer_names[i_stmt_tile],
            outer_tag="g.0",
            inner_tag="l.1",
        )
        t_unit = lp.precompute(  # type: ignore[no-untyped-call]
            t_unit,
            L,
            sweep_inames=[j_inner_name, i_inner_name, new_f],
            precompute_outer_inames=frozenset(
                {j_tile_name, i_tile_name, e_outer_names[i_stmt_tile]}
            ),
            precompute_inames=[i_prftchL, f_prftchL, j_prftchL],
            temporary_name=L_fetch,
            temporary_address_space=lp.AddressSpace.LOCAL,
            default_tag=None,
            compute_insn_id=prefetch_L_insns_ids[i_stmt_tile],
            within=lp_match.Iname(i_inner_name),
        )
        t_unit = lp.split_iname(
            t_unit, i_prftchL, n_e_per_wg, inner_tag="l.1", outer_tag="unr"
        )
        t_unit = lp.split_iname(
            t_unit, j_prftchL, nwork_items_per_e, inner_tag="l.0", outer_tag="unr"
        )

        for istmt, field in enumerate(fields_in_tile):
            subst_name = subst_names[field]
            t_unit = lp.precompute(  # type: ignore[no-untyped-call]
                t_unit,
                subst_name,
                sweep_inames=[
                    e_inner_names[i_stmt_tile],
                    j_inner_names[i_stmt_tile],
                    new_f,
                ],
                temporary_address_space=lp.AddressSpace.LOCAL,
                temporary_name=i_stmt_to_subst_prcmp_tmp[istmt],
                precompute_outer_inames=frozenset(
                    {
                        e_outer_names[i_stmt_tile],
                        j_tile_names[i_stmt_tile],
                    }
                ),
                precompute_inames=[
                    i_stmt_tile_to_e_prcmpt_stage1[i_stmt_tile],
                    i_stmt_tile_to_j_prcmpt_stage1[i_stmt_tile],
                    i_stmt_tile_to_f_prcmpt_stage1[i_stmt_tile],
                ],
                default_tag=None,
            )
        t_unit = lp.tag_inames(
            t_unit, {i_stmt_tile_to_e_prcmpt_stage1[i_stmt_tile]: "l.1"}
        )
        t_unit = lp.split_iname(
            t_unit,
            i_stmt_tile_to_j_prcmpt_stage1[i_stmt_tile],
            nwork_items_per_e,
            inner_tag="l.0",
        )

        t_unit = lp.split_iname(
            t_unit,
            i_inner_name,
            nwork_items_per_e,
            inner_iname=i_inner_inner_name,
            outer_iname=i_inner_outer_name,
            inner_tag="l.0",
            outer_tag="unr",
        )

        outputs_insn_match = lp_match.And(
            (
                within,
                lp_match.Or(
                    tuple(
                        lp_match.Writes(output)
                        for output in i_stmt_tile_to_outputs[i_stmt_tile]
                    )
                ),
            )
        )
        insn_ids = [
            insn.id
            for insn in t_unit[kernel_name].instructions
            if outputs_insn_match(t_unit[kernel_name], insn)
        ]

        t_unit = lp.realize_reduction(t_unit, insn_id_filter=insn_ids)
        inames_to_duplicate = sorted(
            frozenset({i_tile_name, i_inner_outer_name})
            & t_unit[kernel_name].all_inames()
        )
        acc_names = {
            vng(f"acc_{new_f}_{j_tile_name}_{j_inner_name}") for _ in fields_in_tile
        }
        assert (
            set(t_unit[kernel_name].temporary_variables) & acc_names
        ) == acc_names
        t_unit = lp.privatize_temporaries_with_inames(
            t_unit, set(inames_to_duplicate), only_var_names=acc_names
        )
        t_unit = lp.tag_inames(t_unit, {new_f: "unr"})

        new_iname_names_map = {
            i_tile_name: i_stmt_tile_to_itile_init[i_stmt_tile],
            i_inner_outer_name: i_stmt_tile_to_i_inner_outer_init[i_stmt_tile],
        }
        t_unit = lp.duplicate_inames(
            t_unit,
            inames_to_duplicate,
            within=lp_match.Or(
                tuple(
                    lp_match.And(
                        (
                            lp_match.Writes(acc_name),
                            lp_match.Not(lp_match.Reads(acc_name)),
                        )
                    )
                    for acc_name in acc_names
                )
            ),
            new_inames=[new_iname_names_map[iname] for iname in inames_to_duplicate],
            tags=dict.fromkeys(inames_to_duplicate, "unr"),
        )

        new_iname_names_map = {
            i_tile_name: i_stmt_tile_to_itile_assign[i_stmt_tile],
            i_inner_outer_name: i_stmt_tile_to_i_inner_outer_assign[i_stmt_tile],
        }
        t_unit = lp.duplicate_inames(
            t_unit,
            inames_to_duplicate,
            within=lp_match.Or(
                tuple(
                    lp_match.And(
                        (
                            lp_match.Reads(acc_name),
                            lp_match.Not(lp_match.Writes(acc_name)),
                        )
                    )
                    for acc_name in acc_names
                )
            ),
            new_inames=[new_iname_names_map[iname] for iname in inames_to_duplicate],
            tags=dict.fromkeys(inames_to_duplicate, "unr"),
        )

    for i_stmt_tile in range(1, n_stmt_tile):
        predecessor = lp_match.Iname(j_tile_names[i_stmt_tile - 1])
        successor = lp_match.Iname(j_tile_names[i_stmt_tile])
        t_unit = lp.add_dependency(t_unit, successor, predecessor)

    return t_unit


if __name__ == "__main__":
    import os
    from functools import partial

    import pyopencl as cl

    Nfields = 4
    Nface = 4
    Nfacedof = 15
    Nvoldof = 35

    cl_ctx = cl.create_some_context()
    expr = fnsm.batched_einsum(
        "ifj,fe,fej->ei",
        [
            [
                fnsm.array("L", (Nvoldof, Nface, Nfacedof)),
                fnsm.array("J", (Nface, "Nel")),
                fnsm.array(f"v{i}", (Nface, "Nel", Nfacedof)),
            ]
            for i in range(Nfields)
        ],
    )

    if 1:
        fnsm.autotune(expr, os.path.abspath(__file__), cl_ctx)
    else:
        # Enable while debugging ->
        # evaluate a point in the parameter space.
        bound_transform = partial(
            transform,
            n_e_per_wg=16,
            nwork_items_per_e=12,
            n_stmt_tile=2,
            n_i_tile=1,
            n_j_tile=1,
        )

        print(
            fnsm.stringify_comparison_vs_roofline(
                expr, transform=bound_transform, cl_ctx=cl_ctx
            )
        )

# vim: fdm=marker



import logging
import math
from typing import TYPE_CHECKING, Any, cast

import islpy as isl
import loopy as lp
import loopy.match as lp_match
from more_itertools import chunked
from more_itertools import zip_equal as szip

import feinsum as fnsm
import feinsum.loopy_utils as lp_utils
from feinsum.tuning import IntParameter

if TYPE_CHECKING:
    from collections.abc import Collection

logger = logging.getLogger(__name__)


@fnsm.tuning.einsum_arg("noutputs", lambda e: e.b)
@fnsm.tuning.einsum_arg("ndim", lambda e: e.args[0][0].shape[0])
@fnsm.tuning.einsum_arg("ndof", lambda e: e.shape[1])
@fnsm.tuning.transform_param("n_e_per_wg", lambda e: IntParameter(2, 32))
@fnsm.tuning.transform_param(
    "nwork_items_per_e", lambda e: IntParameter(1, e.shape[1])
)
@fnsm.tuning.transform_param("n_stmt_tile", lambda e: IntParameter(1, e.b))
@fnsm.tuning.transform_param(
    "j_tiles", lambda e: IntParameter(1, math.ceil(e.shape[1] / 3))
)
@fnsm.tuning.transform_param(
    "i_tiles", lambda e: IntParameter(1, math.ceil(e.shape[1] / 3))
)
def transform(
    t_unit: lp.TranslationUnit,
    noutputs: int,
    ndim: int,
    ndof: int,
    n_e_per_wg: int,
    nwork_items_per_e: int,
    n_stmt_tile: int,
    i_tiles: int,
    j_tiles: int,
    insn_match: Any | None = None,
    kernel_name: str | None = None,
) -> lp.TranslationUnit:

    if n_e_per_wg * nwork_items_per_e > 600:
        raise fnsm.InvalidParameterError("Block dimension limit exceeded")

    if (
        math.ceil(noutputs / n_stmt_tile)
        * ndim
        * n_e_per_wg
        * math.ceil(ndof / j_tiles)
        + ndim * math.ceil(ndof / i_tiles) * math.ceil(ndof / j_tiles)
    ) * 8e-3 > 47:
        raise fnsm.InvalidParameterError("Shared memory limit exceeded")

    from loopy.match import parse_match

    kernel_name = kernel_name or t_unit.default_entrypoint.name

    within = parse_match(insn_match)
    ref_outputs = ["_fe_out"] + [
        f"_fe_out_{ioutput}" for ioutput in range(noutputs - 1)
    ]

    ref_einsum = fnsm.batched_einsum(
        "xre,rij,xej->ei",
        [
            [
                fnsm.array("J", (ndim, ndim, "Nel")),
                fnsm.array("D", (ndim, ndof, ndof)),
                fnsm.array(f"u{i}", (ndim, "Nel", ndof)),
            ]
            for i in range(noutputs)
        ],
    )

    # {{{ get corresponding variables in t_unit

    vng = t_unit.default_entrypoint.get_var_name_generator()
    ing = t_unit.default_entrypoint.get_instruction_id_generator()
    subst_map = fnsm.match_t_unit_to_einsum(
        t_unit, ref_einsum, insn_match=within, kernel_name=kernel_name
    )
    i = subst_map["i"]
    j = subst_map["j"]
    e = subst_map["e"]
    D = subst_map["D"]
    us = tuple(subst_map[f"u{ioutput}"] for ioutput in range(noutputs))
    outputs = tuple(subst_map[ref_output] for ref_output in ref_outputs)
    x = subst_map["x"]
    r = subst_map["r"]

    # }}}

    # {{{ get new names

    n_stmt_tile = math.ceil(noutputs / math.ceil(noutputs / n_stmt_tile))

    new_is = [vng("i") for _ in range(n_stmt_tile)]
    new_js = [vng("j") for _ in range(n_stmt_tile)]
    new_xs = [vng("x") for _ in range(n_stmt_tile)]
    new_rs = [vng("r") for _ in range(n_stmt_tile)]
    new_es = [vng("e") for _ in range(n_stmt_tile)]
    i_stmt_tile_to_fields = tuple(
        list(el) for el in chunked(us, math.ceil(noutputs / n_stmt_tile))
    )
    i_stmt_tile_to_outputs = tuple(
        list(el) for el in chunked(outputs, math.ceil(noutputs / n_stmt_tile))
    )

    # }}}

    for new_i, new_j, new_x, new_r, new_e, outputs_in_tile in szip(
        new_is, new_js, new_xs, new_rs, new_es, i_stmt_tile_to_outputs
    ):
        t_unit = lp.duplicate_inames(
            t_unit,
            [i, j, x, r, e],
            within=lp_match.And(
                (
                    within,
                    lp_match.Or(
                        tuple(lp_match.Writes(output) for output in outputs_in_tile)
                    ),
                )
            ),
            new_inames=[new_i, new_j, new_x, new_r, new_e],
        )
        t_unit = t_unit.with_kernel(
            lp_utils.decouple_domain(
                t_unit[kernel_name],
                [new_i, new_j, new_x, new_r, new_e],
                parent_inames=cast(
                    "Collection[str]",
                    t_unit[kernel_name]
                    .get_inames_domain(e)
                    .get_var_names(isl.dim_type.param),
                ),
            )
        )

    t_unit = lp.remove_unused_inames(t_unit)

    D_fetch = vng(f"{D}_fetch")
    hoisted_tmp_names = tuple(
        vng("J_times_u_fetch") for _ in i_stmt_tile_to_fields[0]
    )
    D_prftch_id = tuple(ing("D_fetch_insn") for _ in range(n_stmt_tile))

    prcmpt_x_redn_ids = tuple(
        tuple(ing(f"prcmpt_{x}_redn") for _ in fields)
        for fields in i_stmt_tile_to_fields
    )

    for istmt_tile in range(n_stmt_tile):
        i = new_is[istmt_tile]
        j = new_js[istmt_tile]
        x = new_xs[istmt_tile]
        r = new_rs[istmt_tile]
        e = new_es[istmt_tile]
        outputs_in_tile = i_stmt_tile_to_outputs[istmt_tile]
        fields_in_tile = i_stmt_tile_to_fields[istmt_tile]
        subst_names = [vng("hoist_subst") for _ in fields_in_tile]

        knl = t_unit[kernel_name]
        insn_ids = [
            insn.id
            for insn in knl.instructions
            if within(knl, insn)
            and (frozenset(outputs_in_tile) & insn.write_dependency_names())
        ]
        assert len(insn_ids) == len(outputs_in_tile)
        del knl

        # {{{ define variables

        j_inner, j_tile = vng(f"{j}_inner"), vng(f"{j}_tile")
        e_inner, e_outer = vng(f"{e}_inner"), vng(f"{e}_outer")
        i_inner, i_tile = vng(f"{i}_inner"), vng(f"{i}_tile")
        i_inner_inner, i_inner_outer = (
            vng(f"{i_inner}_inner"),
            vng(f"{i_inner}_outer"),
        )
        rprftchD, iprftchD, jprftchD = (
            vng(f"{r}prftchD"),
            vng(f"{i}prftchD"),
            vng(f"{j}prftchD"),
        )
        e_prcmpt_subst, r_prcmpt_subst, j_prcmpt_subst = (
            vng(f"{e}prcmpt_subst"),
            vng(f"{r}prcmpt_subst"),
            vng(f"{j}prcmpt_subst"),
        )

        j_prcmpt_subst_inner, j_prcmpt_subst_outer = (
            vng(f"{j_prcmpt_subst}_inner"),
            vng(f"{j_prcmpt_subst}_outer"),
        )

        # }}}

        # {{{ term hoisting to match the flop count of opt_einsum

        from loopy.symbolic import get_dependencies
        from pymbolic import variables

        knl = t_unit[kernel_name]

        knl = lp.split_reduction_inward(knl, x)
        knl = lp_utils.hoist_invariant_multiplicative_terms_in_sum_reduction(knl, x)
        for subst_name, output in szip(subst_names, outputs_in_tile):
            knl = lp_utils.extract_multiplicative_terms_in_sum_reduction_as_subst(
                knl,
                subst_name=subst_name,
                arguments=variables(f"{e} {j} {r}"),
                within=lp_match.And((within, lp_match.Writes(output))),
                terms_filter=lambda x: (
                    (get_dependencies(x) & knl.all_inames()) <= {r, e, j}
                ),
            )

        knl = lp.split_iname(
            knl,
            j,
            math.ceil(ndof / j_tiles),
            outer_iname=j_tile,
            inner_iname=j_inner,
        )

        t_unit = t_unit.with_kernel(knl)

        # }}}

        t_unit = lp.split_iname(
            t_unit,
            e,
            n_e_per_wg,
            inner_iname=e_inner,
            outer_iname=e_outer,
            inner_tag="l.1",
            outer_tag="g.0",
        )

        # {{{ tile and prefetch D

        t_unit = lp.split_iname(
            t_unit,
            i,
            math.ceil(ndof / i_tiles),
            inner_iname=i_inner,
            outer_iname=i_tile,
            outer_tag="unr",
        )
        t_unit = lp.precompute(  # type: ignore[no-untyped-call]
            t_unit,
            D,
            [i_inner, r, j_inner],
            precompute_outer_inames=frozenset([e_outer, i_tile, j_tile]),
            precompute_inames=[rprftchD, iprftchD, jprftchD],
            temporary_address_space=lp.AddressSpace.LOCAL,
            temporary_name=D_fetch,
            compute_insn_id=D_prftch_id[istmt_tile],
            within=lp_match.And(
                (
                    within,
                    lp_match.Or(
                        tuple(lp_match.Writes(out) for out in outputs_in_tile)
                    ),
                )
            ),
            default_tag=None,
        )
        t_unit = lp.split_iname(t_unit, jprftchD, n_e_per_wg, inner_tag="l.1")
        t_unit = lp.split_iname(t_unit, iprftchD, nwork_items_per_e, inner_tag="l.0")

        # }}}

        # {{{ precompute hoisted substs

        for isubst, (subst_name, tmp_name) in enumerate(
            zip(subst_names, hoisted_tmp_names, strict=False)
        ):
            t_unit = lp.precompute(  # type: ignore[no-untyped-call]
                t_unit,
                subst_name,
                sweep_inames=[r, j_inner, e_inner],
                precompute_inames=[e_prcmpt_subst, j_prcmpt_subst, r_prcmpt_subst],
                precompute_outer_inames=frozenset({e_outer, j_tile}),
                default_tag=None,
                temporary_name=tmp_name,
                compute_insn_id=prcmpt_x_redn_ids[istmt_tile][isubst],
                temporary_address_space=lp.AddressSpace.LOCAL,
            )
        t_unit = lp.tag_inames(t_unit, {e_prcmpt_subst: "l.1"})

        # TODO: It might be worth exploring joining 'r_prcmpt_subst',
        # 'j_prcmpt_subst'.

        t_unit = lp.split_iname(
            t_unit,
            j_prcmpt_subst,
            nwork_items_per_e,
            inner_iname=j_prcmpt_subst_inner,
            outer_iname=j_prcmpt_subst_outer,
            inner_tag="l.0",
            outer_tag="unr",
        )

        # }}}

        t_unit = lp.split_iname(
            t_unit,
            i_inner,
            nwork_items_per_e,
            inner_iname=i_inner_inner,
            outer_iname=i_inner_outer,
            inner_tag="l.0",
            outer_tag="unr",
        )

        for field in fields_in_tile:
            u_fetch = vng("u_fetch")
            t_unit = lp.precompute(  # type: ignore[no-untyped-call]
                t_unit,
                field,
                sweep_inames=[x, j_prcmpt_subst_outer],
                precompute_outer_inames=frozenset(
                    [j_prcmpt_subst_inner, e_prcmpt_subst, e_outer, j_tile]
                ),
                precompute_inames=[vng("prftch_u_x"), None, vng("prftch_u_j")],
                temporary_address_space=lp.AddressSpace.PRIVATE,
                temporary_name=u_fetch,
                default_tag="unr",
            )

        # {{{ TODO: remove once github.com/inducer/loopy/issues/666 is resolved.

        t_unit = lp.realize_reduction(t_unit, insn_id_filter=insn_ids)
        inames_to_duplicate = (
            frozenset({i_tile, i_inner_outer}) & t_unit[kernel_name].all_inames()
        )
        acc_names = set()

        for insn_id in insn_ids:
            (acc_name,) = t_unit[kernel_name].id_to_insn[
                insn_id
            ].read_dependency_names() & set(t_unit[kernel_name].temporary_variables)
            assert acc_name.startswith("acc")
            acc_names.add(acc_name)

        t_unit = lp.privatize_temporaries_with_inames(
            t_unit, inames_to_duplicate, only_var_names=acc_names
        )

        t_unit = lp.duplicate_inames(
            t_unit,
            inames_to_duplicate,
            within=lp_match.Or(
                tuple(
                    lp_match.And(
                        (
                            lp_match.Writes(acc_name),
                            lp_match.Not(lp_match.Reads(acc_name)),
                        )
                    )
                    for acc_name in acc_names
                )
            ),
        )
        t_unit = lp.duplicate_inames(
            t_unit,
            inames_to_duplicate,
            within=lp_match.Or(
                tuple(
                    lp_match.And(
                        (
                            lp_match.Reads(acc_name),
                            lp_match.Not(lp_match.Writes(acc_name)),
                        )
                    )
                    for acc_name in acc_names
                )
            ),
        )

        # }}}

        t_unit = lp.tag_inames(t_unit, {r: "unr", x: "unr"})

    for istmt_tile in range(1, n_stmt_tile):
        predecessor_match = lp_match.Or(
            tuple(
                lp_match.Writes(output)
                for output in i_stmt_tile_to_outputs[istmt_tile - 1]
            )
        )
        for ifield, _ in enumerate(i_stmt_tile_to_fields[istmt_tile]):
            successor_match = lp_match.Or(
                (
                    lp_match.Id(prcmpt_x_redn_ids[istmt_tile][ifield]),
                    lp_match.Id(D_prftch_id[istmt_tile]),
                )
            )

            t_unit = lp.add_dependency(
                t_unit,
                insn_match=successor_match,
                depends_on=lp_match.And((within, predecessor_match)),
            )

    return t_unit


if __name__ == "__main__":
    import os
    from functools import partial

    import pyopencl as cl

    Ndim = 3
    Ndof = 4
    Nfield = 5

    cl_ctx = cl.create_some_context()

    expr = fnsm.batched_einsum(
        "xre,rij,xej->ei",
        [
            [
                fnsm.array("J", (Ndim, Ndim, "Nel")),
                fnsm.array("D", (Ndim, Ndof, Ndof)),
                fnsm.array(f"u{i}", (Ndim, "Nel", Ndof)),
            ]
            for i in range(Nfield)
        ],
    )

    if 1:
        fnsm.autotune(expr, os.path.abspath(__file__), cl_ctx)
    else:
        # Enable while debugging ->
        # evaluate a point in the parameter space.
        bound_transform = partial(
            transform,
            ndim=Ndim,
            ndof=Ndof,
            noutputs=Nfield,
            n_e_per_wg=10,
            nwork_items_per_e=5,
            n_stmt_tile=4,
            i_tiles=15,
            j_tiles=2,
        )

        print(
            fnsm.stringify_comparison_vs_roofline(
                expr, transform=bound_transform, cl_ctx=cl_ctx
            )
        )

# vim: fdm=marker



import json
import sqlite3 as sql

import numpy as np

import feinsum as f


def record_into_db(
    cursor: sql.Cursor,
    *,
    einsum: f.BatchedEinsum,
    device_name: str,
    transform_id: str,
    transform_params: str,
    runtime_in_sec: float,
    compiler_version: str,
    giga_op_info: str,
    timestamp: str,
):
    from feinsum.sql_utils import (
        dump_index_to_length,
        dump_use_matrix,
        dump_value_to_dtype,
    )

    einsum = f.canonicalize_einsum(einsum)
    subscripts = einsum.get_subscripts()
    index_to_length = dump_index_to_length(einsum)
    use_matrix = dump_use_matrix(einsum)
    arg_to_dtype = dump_value_to_dtype(einsum)
    cursor.execute(
        "INSERT INTO FEINSUM_TIMING_FACTS"
        " (subscripts, index_to_length, args,"
        "  arg_to_dtype, device_name, transform_id,"
        "  transform_params, runtime_in_sec,"
        "  compiler_version, giga_op_info, timestamp)"
        " VALUES (?,?,?,?,?,?,?,?,?,?,?)",
        (
            subscripts,
            index_to_length,
            use_matrix,
            arg_to_dtype,
            device_name,
            transform_id,
            transform_params,
            runtime_in_sec,
            compiler_version,
            giga_op_info,
            timestamp,
        ),
    )


def main():
    conn = sql.connect(
        "file:/home/kaushikggg/projects/feinsum/src/"
        "feinsum/data/transform_archive_v5.sqlite?mode=ro",
        uri=True,
    )
    cursor = conn.cursor()
    cursor.execute(
        "SELECT "
        " subscripts,"
        " index_to_length,"
        " use_matrix,"
        " value_to_dtype,"
        " device_name,"
        " transform_id,"
        " transform_params,"
        " runtime_in_sec,"
        " compiler_version,"
        " giga_op_info,"
        " timestamp"
        " FROM FEINSUM_TIMING_FACTS;"
    )
    all_timing_facts = cursor.fetchall()
    conn.close()

    conn = sql.connect("transform_archive_v6.sqlite")
    cursor = conn.cursor()
    cursor.execute(
        "CREATE TABLE FEINSUM_TIMING_FACTS ("
        " ID INTEGER PRIMARY KEY AUTOINCREMENT,"
        " subscripts TEXT,"
        " index_to_length TEXT,"
        " args TEXT,"
        " arg_to_dtype TEXT,"
        " device_name TEXT,"
        " transform_id TEXT,"
        " transform_params TEXT,"
        " runtime_in_sec REAL,"
        " compiler_version TEXT,"
        " giga_op_info TEXT,"
        " timestamp TEXT"
        ")"
    )

    print("Number of timing facts = ", len(all_timing_facts))
    for fact in all_timing_facts:
        (
            subscripts,
            idx_to_len,
            use_matrix,
            val_to_dtype,
            device_name,
            transform_id,
            transform_params,
            runtime_in_sec,
            compiler_version,
            giga_op_info,
            timestamp,
        ) = fact
        input_subscripts = subscripts.split("->")[0]
        idx_to_len = json.loads(idx_to_len)
        val_to_dtype = {
            val: np.dtype(dtype) for val, dtype in json.loads(val_to_dtype).items()
        }
        assert all(
            [all(len(uses) == 1 for uses in use_row)]
            for use_row in json.loads(use_matrix)
        )
        arg_names = [
            [uses[0] for uses in use_row]
            for use_row in json.loads(use_matrix)
        ]
        arg_shapes = [
            [
                idx_to_len.get(idx, idx.upper())
                for idx in input_subscript
            ]
            for input_subscript in input_subscripts.split(",")
        ]
        args = [
            [
                f.array(arg_name, arg_shape, val_to_dtype[arg_name])
                for arg_name, arg_shape in zip(arg_row, arg_shapes, strict=True)
            ]
            for arg_row in arg_names
        ]
        einsum = f.batched_einsum(subscripts, args)
        record_into_db(
            cursor,
            einsum=einsum,
            device_name=device_name,
            transform_id=transform_id,
            transform_params=transform_params,
            runtime_in_sec=runtime_in_sec,
            compiler_version=compiler_version,
            giga_op_info=giga_op_info,
            timestamp=timestamp,
        )
        print("Recording a fact.")

    conn.commit()
    conn.close()


if __name__ == "__main__":
    main()
